"""
Scripting interface for GenX for use in python scripting or Jupyter Notebooks.
"""

__all__=[]

import os
import sys

# workaround for issues with ctrl+c on windows
if sys.platform=='win32':
    os.environ['FOR_DISABLE_CONSOLE_CTRL_HANDLER']='1'

from .model import Model
from .diffev import DiffEv
from .model_control import ModelController
from .plugins.add_ons.help_modules.reflectivity_utils import avail_models, SampleHandler, SampleBuilder
from .plugins.utils import PluginHandler

_fit_output=[]

controller=ModelController(DiffEv())

def text_output_api(text):
    _fit_output.append(text)

def load(fname, compile=True):
    controller.load_file(fname)
    if compile:
        controller.model.compile_script()
    controller.optimizer.model=controller.model
    return controller.model, controller.optimizer

def save(fname, model=None, optimizer=None):
    if model is not None:
        controller.model=model
    if optimizer is not None:
        controller.optimizer=optimizer
    controller.save_file(fname)

def fit_notebook(model, optimizer):
    """
    Function to fit a GenX model while giving feedback on a Jupyter notebook.
    """
    global _fit_output
    _fit_output=[]
    optimizer.text_output=text_output_api
    optimizer.start_fit(model)
    import matplotlib.pyplot as plt
    from IPython.display import display, clear_output
    from numpy import array
    from time import sleep

    fig=plt.figure(figsize=(14, 5))
    plt.suptitle("To stop fit, interrupt the Kernel")

    plt.subplot(121)
    ax1=plt.gca()
    line=plt.semilogy([0., 1.], [0.1, 1.])[0]
    plt.xlabel('Generation')
    plt.ylabel('FOM')
    t1=plt.title('FOM:')

    plt.subplot(122)
    t2=plt.title('Data display')
    ax2=plt.gca()
    refls=[]
    for i, ds in enumerate(model.data.items):
        refls.append(plt.semilogy(ds.x, ds.y,
                                  color=ds.data_color,
                                  lw=ds.data_linethickness, ls=ds.data_linetype,
                                  marker=ds.data_symbol, ms=ds.data_symbolsize,
                                  label='data-%i: %s'%(i, ds.name))[0])
        if ds.y_sim.shape==ds.y.shape:
            refls.append(plt.semilogy(ds.x, ds.y_sim,
                                      color=ds.sim_color,
                                      lw=ds.sim_linethickness, ls=ds.sim_linetype,
                                      marker=ds.sim_symbol, ms=ds.sim_symbolsize,
                                      label='model-%i: %s'%(i, ds.name))[0])

    plt.xlabel('x')
    plt.ylabel('I')

    plt.draw()

    display(fig)

    last=2
    while optimizer.running:
        try:
            sleep(0.1)
            if len(optimizer.fom_log)<=last:
                continue
            x, y=array(optimizer.fom_log).T
            last=len(x)
            # t1.set_text('FOM: %.4e'%optimizer.best_fom)
            t1.set_text(_fit_output[-1])
            # vec=optimizer.best_vec
            # list(map(lambda func, value: func(value), model.get_fit_pars()[0], vec))
            # model.evaluate_sim_func()
            ax2.clear()
            t2 = plt.title('Data display')
            refls = []
            for i, ds in enumerate(model.data.items):
                refls.append(plt.semilogy(ds.x, ds.y,
                                          color=ds.data_color,
                                          lw=ds.data_linethickness, ls=ds.data_linetype,
                                          marker=ds.data_symbol, ms=ds.data_symbolsize,
                                          label='data-%i: %s'%(i, ds.name))[0])
                if ds.y_sim.shape==ds.y.shape:
                    refls.append(plt.semilogy(ds.x, ds.y_sim,
                                              color=ds.sim_color,
                                              lw=ds.sim_linethickness, ls=ds.sim_linetype,
                                              marker=ds.sim_symbol, ms=ds.sim_symbolsize,
                                              label='model-%i: %s'%(i, ds.name))[0])

            plt.xlabel('x')
            plt.ylabel('I')

            line.set_xdata(x)
            line.set_ydata(y)

            ax1.set_xlim(0, x[-1])
            ax1.set_ylim(y.min()*0.9, y.max()*1.1)
            try:
                plt.draw()
            except Exception as e:
                print(e)
            else:
                clear_output(wait=True)
                display(fig)
        except KeyboardInterrupt:
            optimizer.stop_fit()
    plt.close()

    print(_fit_output[-1])
    print("If you want to update the model with the fit results, call api.fit_update(model, optimizer)")

def fit(model, optimizer):
    """
    Function to fit a GenX model while giving feedback with matplotlib graphs.
    """
    optimizer.text_output=print
    optimizer.start_fit(model)
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        plot_result=False
    else:
        plot_result=True
        from numpy import array
        from time import sleep

    if plot_result:
        fig=plt.figure(figsize=(14, 5))
        plt.ion()
        plt.suptitle("To stop fit, close the figure window")

        plt.subplot(121)
        ax1=plt.gca()
        line=plt.semilogy([0., 1.], [0.1, 1.])[0]
        plt.xlabel('Generation')
        plt.ylabel('FOM')
        t1=plt.title('FOM:')

        plt.subplot(122)
        t2=plt.title('Data display')
        ax2=plt.gca()
        refls=[]
        for i, ds in enumerate(model.data.items):
            refls.append(plt.semilogy(ds.x, ds.y,
                                      color=ds.data_color,
                                      lw=ds.data_linethickness, ls=ds.data_linetype,
                                      marker=ds.data_symbol, ms=ds.data_symbolsize,
                                      label='data-%i: %s'%(i, ds.name))[0])
            if ds.y_sim.shape==ds.y.shape:
                refls.append(plt.semilogy(ds.x, ds.y_sim,
                                          color=ds.sim_color,
                                          lw=ds.sim_linethickness, ls=ds.sim_linetype,
                                          marker=ds.sim_symbol, ms=ds.sim_symbolsize,
                                          label='model-%i: %s'%(i, ds.name))[0])

        plt.xlabel('x')
        plt.ylabel('I')

        fig.canvas.mpl_connect('close_event', lambda event: optimizer.stop_fit())
        plt.draw()

    last=2
    print("To stop fit, press ctrl+C")
    while optimizer.running:
        try:
            if plot_result:
                if len(optimizer.fom_log)<=last:
                    plt.pause(0.1)
                    continue
                x, y=array(optimizer.fom_log).T
                last=len(x)
                j=0
                for i, ds in enumerate(model.data.items):
                    refls[j].set_ydata(ds.y)
                    if ds.y_sim.shape==ds.y.shape:
                        j=+1
                        refls[j].set_ydata(ds.y_sim)
                    j+=1

                line.set_xdata(x)
                line.set_ydata(y)

                ax1.set_xlim(0, x[-1])
                ax1.set_ylim(y.min()*0.9, y.max()*1.1)
                plt.draw()
            else:
                sleep(0.1)
        except KeyboardInterrupt:
            optimizer.stop_fit()
    if plot_result:
        plt.close()
    print("If you want to update the model with the fit results, call api.fit_update(model, optimizer)")

def fit_update(model, optimizer):
    pnames=model.parameters.get_fit_pars()[1]
    for di in model.parameters.data:
        if di[0] in pnames:
            di[1]=optimizer.best_vec[pnames.index(di[0])]
    model.simulate()

class Reflectivity(SampleBuilder):
    """
    Interface to build a model script for reflectivity simulations. Surves the same purpose
    as the Reflectivity plugin in the GUI.
    """
    _uvar_string='cp.new_var('

    def __init__(self, model, optimizer, analyze_model=True):
        SampleBuilder.__init__(self, model)
        self.optimizer=optimizer
        if analyze_model:
            self.ReadModel()

    @classmethod
    def create_new(cls, modelname='spec_nx'):
        """
        Create a completely new GenX model for reflectivity.

        returns model, optimizer, Reflectivity
        """
        model=Model()
        optimizer=DiffEv()
        optimizer.model=model
        out=cls(model, optimizer, analyze_model=False)
        out.new_model(modelname)
        return model, optimizer, out

    def AppendSim(self, sim_func, inst, args):
        self.sim_funcs.append(sim_func)
        self.sim_insts.append(inst)
        self.sim_args.append(args)

    def new_model(self, modelname='spec_nx'):
        '''Init the script in the model to yield the
        correct script for initilization
        '''
        modelname='models.'+modelname
        model_data=self.GetModel().get_data()
        nb_data_sets=len(model_data)

        script=self.GetNewModelScript(modelname, nb_data_sets)
        self.BuildNewModel(script)

        self.instruments={'inst': self.model.Instrument()}

        self.data_names=[data_set.name for data_set in model_data]
        self.uservars_lines=[]
        self.sim_exp=[[] for item in self.data_names]
        self.sim_funcs=['Specular']*nb_data_sets
        self.sim_insts=['inst']*nb_data_sets
        self.sim_args=[['d.x'] for i in range(nb_data_sets)]
        self.WriteModel()

    def WriteModel(self):
        parameter_list=self.uservars_lines
        sim_funcs, sim_insts, sim_args=self.sim_funcs, self.sim_insts, self.sim_args
        expression_list=self.sim_exp
        instruments=self.instruments

        # make sure the number of datasets is reflected correctly
        model_data=self.GetModel().get_data()
        nb_data_sets=len(model_data)
        diff=nb_data_sets-len(sim_funcs)
        if diff>0:
            expression_list+=[[] for item in range(diff)]
            for i in range(diff):
                self.insert_new_data_segment(len(sim_funcs))
        elif diff<0:
            for i in range(-diff):
                expression_list.pop(-1)
                self.remove_data_segment(nb_data_sets)

        self.write_model_script(sim_funcs, sim_insts, sim_args,
                                expression_list, parameter_list, instruments)

    def update_script(self):
        self.WriteModel()

    def add_data(self, name=''):
        self.GetModel().data.add_new(name=name)
        self.update_script()

    def ReadModel(self):
        '''ReadModel(self)  --> None

        Reads in the current model and locates layers and stacks
        and sample defined inside BEGIN Sample section.
        '''
        self.CompileScript()

        instrument_names=self.find_instrument_names()

        if len(instrument_names)==0:
            raise ValueError('Could not find any Instruments in the'
                             ' model script. Check the script.')

        if not 'inst' in instrument_names:
            raise ValueError('Could not find the default Instrument, inst, in the'
                             ' model script. Check the script.')

        sample_text=self.find_sample_section()

        if sample_text is None:
            raise ValueError('Could not find the sample section in the model script.\n '
                             'Can not load the sample in the editor.')

        all_names, layers, stacks=self.find_layers_stacks(sample_text)

        if len(layers)==0:
            raise ValueError('Could not find any Layers in the'
                             ' model script. Check the script.')

        # Now its time to set all the parameters so that we have the strings
        # instead of the evaluated value - looks better
        for lay in layers:
            for par in lay[1].split(','):
                vars=par.split('=')
                exec('%s.%s = "%s"'%(lay[0], vars[0].strip(), vars[1].strip()), self.GetModel().script_module.__dict__)

        data_names, insts, sim_args, sim_exp, sim_funcs=self.find_sim_function_parameters()

        uservars_lines=self.find_user_parameters()

        self.model=self.GetModel().script_module.model
        sample=self.GetModel().script_module.sample

        self.sampleh=SampleHandler(sample, all_names)
        self.sampleh.model=self.model
        instruments={}
        for name in instrument_names:
            instruments[name]=getattr(self.GetModel().script_module, name)
        self.instruments=instruments

        self.data_names=data_names
        self.sim_exp=sim_exp
        self.uservars_lines=uservars_lines

        self.sim_funcs=sim_funcs
        self.sim_insts=insts
        self.sim_args=sim_args

        self.GetModel().compiled=False

    def add_stack(self, name, stack=None, **kwargs):
        if self.sampleh.checkName(name) or not name.isidentifier():
            raise KeyError("Name already exists or not proper identifier")
        self.sampleh.insertItem(len(self.sampleh)-1, 'Stack', name)
        stack=self.sampleh[name]
        if stack is not None:
            self[name]=stack
        else:
            for key, value in kwargs.items():
                setattr(stack, key, value)
            self.WriteModel()
        return self[name]

    def add_layer(self, name, layer=None, **kwargs):
        if self.sampleh.checkName(name) or name in self.instruments or not name.isidentifier():
            raise KeyError("Name already exists or not proper identifier")
        self.sampleh.insertItem(len(self.sampleh)-1, 'Layer', name)
        lay=self.sampleh[name]
        if layer is not None:
            self[name]=layer
        else:
            for key, value in kwargs.items():
                setattr(lay, key, value)
            self.WriteModel()
        return self[name]

    def add_instrument(self, name, instrument=None, **kwargs):
        if self.sampleh.checkName(name) or name in self.instruments or not name.isidentifier():
            raise KeyError("Name already exists or not proper identifier")
        if instrument is None:
            self.instruments[name]=self.Instrument(**kwargs)
            self.WriteModel()
        else:
            self.instruments[name]=instrument
        return self[name]

    def move_up(self, item):
        self.sampleh.moveUp(item)
        self.WriteModel()

    def move_down(self, item):
        self.sampleh.move_down(item)
        self.WriteModel()

    @property
    def Layer(self):
        return self.model.Layer

    @property
    def Stack(self):
        return self.model.Stack

    @property
    def Instrument(self):
        return self.model.Instrument

    def __getitem__(self, item):
        if item in self.sampleh.names:
            return self.sampleh[item]
        elif item in self.instruments:
            return self.instruments[item]
        else:
            raise KeyError("Has to be name of a Stack, Layer or Instrument")

    def __setitem__(self, key, value):
        if key in self.sampleh.names:
            prev=self.sampleh[key]
            if type(prev)!=type(value):
                raise ValueError("Has to be of type %s"%type(prev))
            for k in prev._parameters.keys():
                setattr(prev, k, getattr(value, k))
        elif key in self.instruments:
            prev=self.instruments[key]
            if type(prev)!=type(value):
                raise ValueError("Has to be of type %s"%type(prev))
            self.instruments[key]=value
        else:
            raise KeyError("Has to be name of a Stack, Layer or Instrument")
        self.WriteModel()

    def __repr__(self):
        output='Reflectivity Model:\n'
        output+='    Instruments:'
        for key, value in self.instruments.items():
            output+='\n        %s = %s'%(key, value)

        output+='\n    Sample Structure:\n        '
        output+='\n        '.join(self.sampleh.getStringList(html_encoding=False))

        output+='\n    User Parameters:'
        for line in self.uservars_lines:
            var, val=line[len(self._uvar_string):].split(')')[0].split(',')
            output+='\n        %s = %s'%(eval(var), val)

        output+='\n    Simulations:'
        for name, func, inst, exps in zip(self.data_names, self.sim_funcs, self.sim_insts, self.sim_exp):
            output+='\n        %s: %s(%s, %s)'%(name, func, 'd.x', inst)
            for ei in exps:
                output+='\n            %s'%ei

        return output

    def _repr_html_(self):
        output='<h3>Reflectivity Model</h3>'
        output+='<h4>Instruments:</h4>\n<div>'
        for key, value in self.instruments.items():
            output+=self.sampleh.htmlize('%s = %s'%(key, value))
            output+='<br />'
        output+='</div>'

        output+='<h4>Sample Structure:</h4>\n<div>'
        output+='<br />\n'.join(self.sampleh.getStringList(html_encoding=True))
        output+='</div>'

        output+='<h4>User Parameters:</h4>\n<div><ul>'
        for line in self.uservars_lines:
            var, val=line[len(self._uvar_string):].split(')')[0].split(',')
            output+='<li><b>%s</b> = %s</li>'%(eval(var), val)
        output+='</ul></div>'

        output+='<h4>Simulations:</h4>\n<div>'
        for name, func, inst, exps in zip(self.data_names, self.sim_funcs, self.sim_insts, self.sim_exp):
            output+='<b>%s:</b> %s(%s, %s)'%(name, func, 'd.x', inst)
            output+='<ul>'
            for ei in exps:
                output+='<li>%s</li>'%ei
            output+='</ul><br />'
        output+='</div>'

        return output

    @property
    def widget(self):
        # alternative widget to the model view with reflectivity controlls
        import ipywidgets as ipw
        model=self.GetModel()
        try:
            model.simulate()
        except Exception as error:
            print(error)

        graphw=ipw.Output()
        with graphw:
            from matplotlib import pyplot as plt
            fig=plt.figure(figsize=(10, 8))
            model.data.plot()
            plt.xlabel('q/tth')
            plt.ylabel('Intensity')
            from IPython.display import display
            display(fig)
            plt.close()

        dataw=model.data._repr_ipyw_()
        top=ipw.HBox([dataw, graphw])
        parameters=model.parameters._repr_ipyw_()

        replot=ipw.Button(description='simulate')
        replot.on_click(model._ipyw_replot)
        replot._plot_output=graphw

        tabs=ipw.Tab(children=[parameters, self.sample_widget, self.instrument_widget])
        tabs.set_title(0, 'Parameters')
        tabs.set_title(1, 'Sample')
        tabs.set_title(2, 'Instrument')

        return ipw.VBox([replot, top, tabs])

    @property
    def instrument_widget(self):
        import ipywidgets as ipw
        entries=[ipw.HTML('<b>Instruments (use string to define formulas):</b>')]
        for key, value in self.instruments.items():
            ient=[]
            if key=='inst':
                name=ipw.HTML(key, layout=ipw.Layout(width='14ex'))
            else:
                name=ipw.Text(value=key, layout=ipw.Layout(width='14ex'))
                name.observe(self._ipyw_change_iname, names='value')
            ient.append(name)
            ient.append(ipw.HTML('=Instrument(', layout=ipw.Layout(width='12ex')))

            items=str(value).split('(', 1)[1].rsplit(')', 1)[0].split(',')
            for item in items:
                try:
                    name, val=item.split('=', 1)
                except ValueError:
                    break
                name=name.strip()
                label_item=ipw.HTML('%s='%name, layout=ipw.Layout(width='%iex'%(len(name)+1)))
                ient.append(label_item)
                entry=ipw.Text(value=val, layout=ipw.Layout(width='18ex'))
                ient.append(entry)
                entry.change_item=name
                entry.change_name=key
                entry.label_item=label_item
                entry.observe(self._ipyw_change_eval, names='value')

            ient.append(ipw.HTML(')', layout=ipw.Layout(width='1ex')))
            entries.append(ipw.HBox(ient))
        btn=ipw.Button(description='Add Instrument')
        btn.on_click(self._ipyw_new_instrument)
        entries.append(btn)
        vbox=ipw.VBox(entries)
        btn.vbox=vbox
        return vbox

    @property
    def sample_widget(self):
        import ipywidgets as ipw
        entries=[ipw.HTML('<b>Sample (use string to define formulas):</b>')]

        maxlen=200
        vbox=ipw.VBox([])

        for i, line in enumerate(self.sampleh.getStringList(html_encoding=False)):
            key, data=map(str.strip, line.split('=', 1))
            ient=[]
            if key in ['Amb', 'Sub']:
                name_entr=ipw.HTML(key, layout=ipw.Layout(width='14ex'))
            else:
                name_entr=ipw.Text(value=key, layout=ipw.Layout(width='14ex'))
                name_entr.observe(self._ipyw_change_sname, names='value')
            ient.append(name_entr)

            if data.startswith('model.Stack'):
                ient.append(ipw.HTML('=Stack(', layout=ipw.Layout(width='12ex')))
                values=data[12:].strip()
            else:
                if key not in ['Amb', 'Sub']:
                    ient.insert(0, ipw.HTML("    ", layout=ipw.Layout(width='4ex')))
                cls, values=map(str.strip, data.split('(', 1))
                values=values.rsplit(')', 1)[0]
                ient.append(ipw.HTML('=%s('%cls[6:], layout=ipw.Layout(width='12ex')))

            thislen=14+12+4
            items=values.split(',')
            for item in items:
                try:
                    name, val=item.split('=', 1)
                except ValueError:
                    break
                name=name.strip()
                label_item=ipw.HTML('%s='%name, layout=ipw.Layout(width='%iex'%(len(name)+1)))
                ient.append(label_item)
                entry=ipw.Text(value=val, layout=ipw.Layout(width='18ex'))
                ient.append(entry)
                entry.change_item=name
                entry.change_name=key
                entry.label_item=label_item
                entry.observe(self._ipyw_change, names='value')
                thislen+=20+(len(name)+1)

            maxlen=max(maxlen, thislen+1)
            ient.append(ipw.HTML(')', layout=ipw.Layout(width='1ex')))
            if data.startswith('model.Stack'):
                btn=ipw.Button(description='Add Layer')
                btn.on_click(self._ipyw_new_layer)
                btn.change_stack=key
                btn.vbox=vbox
                ient.append(btn)
            else:
                name_entr.input_widgets=tuple(wi for wi in ient[2:] if type(wi) is ipw.Text)
            entries.append(ipw.HBox(ient))

        btn=ipw.Button(description='Add Stack')
        btn.on_click(self._ipyw_new_stack)
        entries.append(btn)
        btn.vbox=vbox
        vbox.children=tuple(entries)
        vbox.layout.width='%iex'%maxlen
        return vbox

    def _ipyw_change(self, change):
        item=self[change.owner.change_name]
        try:
            eval(change.new, self.GetModel().script_module.__dict__)
        except Exception as err:
            change.owner.label_item.value='<div style="color: red;">%s=</div>'%change.owner.change_item
        else:
            change.owner.label_item.value='%s='%change.owner.change_item
            setattr(item, change.owner.change_item, change.new)
            self.WriteModel()

    def _ipyw_change_eval(self, change):
        item=self[change.owner.change_name]
        try:
            setattr(item, change.owner.change_item, eval(change.new, self.GetModel().script_module.__dict__))
        except Exception as err:
            change.owner.label_item.value='<div style="color: red;">%s=</div>'%change.owner.change_item
        else:
            change.owner.label_item.value='%s='%change.owner.change_item
            self.WriteModel()

    def _ipyw_change_iname(self, change):
        self.instruments[change.new]=self.instruments[change.old]
        del (self.instruments[change.old])
        for i, inst in enumerate(self.sim_insts):
            if inst==change.old:
                self.sim_insts[i]=str(change.new)
        self.WriteModel()

    def _ipyw_new_instrument(self, btn):
        key='inst_%i'%len(self.instruments)
        self.add_instrument(key)
        value=self[key]

        import ipywidgets as ipw
        ient=[]
        name=ipw.Text(value=key, layout=ipw.Layout(width='14ex'))
        name.observe(self._ipyw_change_iname, names='value')
        ient.append(name)
        ient.append(ipw.HTML('=Instrument(', layout=ipw.Layout(width='12ex')))

        items=str(value).split('(', 1)[1].rsplit(')', 1)[0].split(',')
        for item in items:
            try:
                name, val=item.split('=', 1)
            except ValueError:
                break
            name=name.strip()
            label_item=ipw.HTML('%s='%name, layout=ipw.Layout(width='%iex'%(len(name)+1)))
            ient.append(label_item)
            entry=ipw.Text(value=val, layout=ipw.Layout(width='18ex'))
            ient.append(entry)
            entry.change_item=name
            entry.change_name=key
            entry.label_item=label_item
            entry.observe(self._ipyw_change, names='value')

        ient.append(ipw.HTML(')', layout=ipw.Layout(width='1ex')))
        btn.vbox.children=btn.vbox.children[:-1]+(ipw.HBox(ient), btn.vbox.children[-1])
        self.WriteModel()

    def _ipyw_new_layer(self, btn):
        idx=[bi.children[-1] for bi in btn.vbox.children[1:-1]].index(btn)+1
        key='new_layer_%i'%len(self.sampleh.names)
        self.sampleh.insertItem(btn.change_stack, 'Layer', key)
        lay=self[key]

        import ipywidgets as ipw
        ient=[]
        name_entr=ipw.Text(value=key, layout=ipw.Layout(width='14ex'))
        name_entr.observe(self._ipyw_change_sname, names='value')
        ient.append(name_entr)
        data=repr(lay)

        ient.insert(0, ipw.HTML("    ", layout=ipw.Layout(width='4ex')))
        cls, values=map(str.strip, data.split('(', 1))
        values=values.rsplit(')', 1)[0]
        ient.append(ipw.HTML('=%s('%cls, layout=ipw.Layout(width='12ex')))

        items=values.split(',')
        for item in items:
            try:
                name, val=item.split('=', 1)
            except ValueError:
                break
            name=name.strip()
            label_item=ipw.HTML('%s='%name, layout=ipw.Layout(width='%iex'%(len(name)+1)))
            ient.append(label_item)
            entry=ipw.Text(value=val, layout=ipw.Layout(width='18ex'))
            ient.append(entry)
            entry.change_item=name
            entry.change_name=key
            entry.label_item=label_item
            entry.observe(self._ipyw_change, names='value')
        ient.append(ipw.HTML(')', layout=ipw.Layout(width='1ex')))
        name_entr.input_widgets=tuple(wi for wi in ient[2:] if type(wi) is ipw.Text)

        btn.vbox.children=btn.vbox.children[:idx+1]+(ipw.HBox(ient),)+btn.vbox.children[idx+1:]
        self.WriteModel()

    def _ipyw_change_sname(self, change):
        idx=self.sampleh.names.index(change.old)
        self.sampleh.names[idx]=change.new
        # update the name for the layer entries
        for entr in change.owner.input_widgets:
            entr.change_name=change.new
        self.WriteModel()

    def _ipyw_new_stack(self, btn):
        pass

class DataLoaderInterface():

    def __init__(self):
        head=os.path.dirname(os.path.abspath(__file__))
        self._handler=PluginHandler(None, os.path.join(head, 'plugins', ''), 'data_loaders')
        for dl in self._handler.get_possible_plugins():
            try:
                self._handler.load_plugin(dl)
                setattr(self, dl, self._handler.loaded_plugins[dl])
            except Exception as error:
                print(error)

    def __repr__(self):
        output="Available data loaders:"
        for dl in sorted(self._handler.loaded_plugins.keys()):
            output+='\n  data_loader.%s.LoadData(dataset, filename)'%dl
        return output

data_loader=DataLoaderInterface()



"""
Module to define all user exceptions raised by GenX.
"""

class GenxError(Exception):
    '''Basic GenX error class to allow filtering for all specific GenX errors.'''
    pass

class GenxIOError(IOError, GenxError):
    ''' Error class for input output, mostly concerning files'''

    def __init__(self, error_message, file=''):
        self.error_message=error_message
        self.file=file
        text='Input/Output error for file:\n'+self.file+ \
             '\n\n Python error:\n '+self.error_message
        IOError.__init__(self, text)

class GenxOptionError(IndexError, GenxError):
    ''' Error class for not finding an option section pair in the configuration '''

    def __init__(self, section, option):
        self.section=section
        self.option=option
        text='Error in trying to loacate values in GenX configuration.'+ \
             '\nCould not locate the section: '+self.section+ \
             ' or option: '+self.option+'.'
        IndexError.__init__(self, text)

class ErrorBarError(RuntimeError, GenxError):
    def __init__(self, error_message=None):
        text=error_message
        if error_message is None:
            text='Could not evaluate the error bars. A fit has to be run before they can be calculated'
        RuntimeError.__init__(self, text)

class ParameterError(ValueError, GenxError):
    ''' Class for yielding Parameter errors
    '''
    def __init__(self, parameter: str, parameter_number: int, py_error: Exception=None, what:int=-1):
        '''
        parameter: the name of the parameter
        parameter_number: the position of the parameter in the list
        error_mesage: pythons error from the original exception
        int: where the error occured
            -1 : undefined
             0 : an not find the parameter
             1 : can not evaluate i.e. set the parameter
             2 : value are larger than max
             3 : value are smaller than min
             4 : No parameters to fit
        '''
        self.parameter=parameter
        self.parameter_number=parameter_number
        self.what=what

        text=''
        text+='Parameter number %i, %s, '%(self.parameter_number,
                                           self.parameter)

        # Take care of the different cases
        if self.what==0:
            text+='could not be found. Check the spelling.\n'
        elif self.what==1:
            text+='could not be evaluated. Check the code of the function.\n'
        elif self.what==2:
            text+='is larger than the value in the max column.\n'
        elif self.what==3:
            text+='is smaller than the value in the min column\n'
        elif self.what==4:
            text='There are no parameter selcted to be fitted.\n'+ \
                 'Select the parameters you want to fit by checking the '+ \
                 'boxes in the fit column, folder grid'
        else:
            text+='yielded an undefined error. Check the Python output\n'

        if py_error is not None:
            text+='\nPython error output:\n'+str(py_error)
        ValueError.__init__(self, text)

class ModelError(RuntimeError, GenxError):
    ''' Class for yielding compile or evaluation errors in the model text
    '''

    def __init__(self, error_message, where):
        '''
        error_mesage: pythons error message from the original exception
        where: integer describing where the error was raised.
                -1: undef
                 0: compile error
                 1: evaulation error
        '''
        self.where=where
        text=''
        if self.where==0:
            text+='It was not possible to compile the model script.\n'
        elif self.where==1:
            text+='It was not possible to evaluate the model script.\n' \
                  +'Check the Sim function.\n'
        elif self.where==-1:
            text+='Undefined error from the Model. See below.\n'
        text+='\n'+error_message
        RuntimeError.__init__(self, text)

class FomError(RuntimeError, GenxError):
    '''Error class for the fom evaluation'''

    def __init__(self, error_message):
        text='Could not evaluate the FOM function. See python output.\n' \
             +'\n'+error_message
        RuntimeError(self, text)

class OptimizerInterrupted(RuntimeError, GenxError):
    '''Error raised to stop a running refinement function from within a thread'''


#!/usr/bin/env python
import logging
import multiprocessing
from threading import Thread

import appdirs
import argparse
import os
import os.path
import sys
from logging import debug, getLogger

from . import version
from .core import custom_logging
from .core.custom_logging import activate_excepthook, activate_logging, iprint, setup_system


def start_interactive(args):
    '''
    Start genx in interactive mode (with the gui)

    :param args: command line arguments evaluated with argparse.
    '''
    debug('enter start_interactive')
    activate_excepthook()
    # Fix blurry text on Windows 10
    import ctypes
    try:
        ctypes.windll.shcore.SetProcessDpiAwareness(True)
    except:
        pass

    from .gui import main_window
    if args.infile!='':
        debug('start GUI setup with file to load')
        filename = args.infile
    else:
        filename = None
    debug('start GUI setup')
    app = main_window.GenxApp(filename=filename, dpi_overwrite=args.dpi_overwrite)
    debug('setup complete, start WX MainLoop')
    app.MainLoop()
    debug('leave start_interactive')


def create_simulated_data(args):
    """Function to create simulated data from the model and add it to data.y"""
    from .diffev import DiffEv
    from .model_control import ModelController
    from .core import config as io

    from scipy.stats import poisson

    io.config.load_default(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'profiles', 'default.profile'))
    ctrl = ModelController(DiffEv())
    mod = ctrl.model

    iprint("Loading file: %s "%args.infile)
    ctrl.load_file(args.infile)
    # io.load_opt_config(opt, config)
    iprint("File loaded")

    iprint("Simualting..")
    mod.simulate()
    iprint("Storing simualted data")
    for data_set in mod.data:
        data_set.y_raw = poisson.rvs(data_set.y_sim)
        data_set.y_command = 'y'
        data_set.run_y_command()
        data_set.error_command = 'sqrt(where(y > 0, y, 1.0))'
        data_set.run_error_command()

    iprint('Saving the model to %s'%args.outfile)
    ctrl.save_file(args.outfile)


def extract_parameters(args):
    """Extracts the parameters to outfile"""

    from .diffev import DiffEv
    from .model_control import ModelController
    from .core import config as io

    # Open the genx file
    io.config.load_default(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'profiles', 'default.profile'))
    ctrl = ModelController(DiffEv())
    mod = ctrl.model

    iprint("Loading file: %s "%args.infile)
    ctrl.load_file(args.infile)
    iprint("File loaded")

    names, values = mod.parameters.get_sim_pars()
    index_names = mod.parameters.get_names()
    errors = [mod.parameters[index_names.index(ni)].error if mod.parameters.get_fit_state_by_name(ni)==1
              else '-' for ni in names]
    names = [ni if mod.parameters.get_fit_state_by_name(ni)!=1 else ni+'*' for ni in names]

    paramstr = '\t'.join([name for name in names])+'\n'+ \
               '\t'.join(['%f'%val for val in values])+'\n'
    if args.error:
        paramstr += '\t'.join([err for err in errors])+'\n'
    if args.outfile!='':
        iprint(f'Appending parameter values to file {args.outfile}')
        with open(args.outfile, 'a') as fout:
            # Add header
            fout.write(paramstr)
    iprint('\n'+paramstr)


def modify_file(args):
    """Modify a GenX file given command line arguments"""
    from .diffev import DiffEv
    from .model_control import ModelController
    from .core import config as io

    # Open the genx file
    io.config.load_default(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'profiles', 'default.profile'))
    ctrl = ModelController(DiffEv())
    mod = ctrl.model

    iprint("Loading file: %s "%args.infile)
    ctrl.load_file(args.infile)
    iprint("File loaded")

    if args.datafile:
        datafile = os.path.abspath(args.datafile)
        if 0>args.data_set or args.data_set>=len(mod.data):
            iprint("The selected data set does not exist - select one between 0 and %d"%(len(mod.data)-1))
            return
        iprint('Loading dataset %s into data set %d'%(datafile, args.data_set))
        mod.data[args.data_set].loadfile(datafile)

        if args.outfile:
            iprint('Saving the fit to %s'%args.outfile)
            ctrl.save_file(args.outfile)

    elif args.save_datafile:
        save_datafile = os.path.abspath(args.save_datafile)
        iprint("Simualting..")
        mod.simulate()
        if 0>args.data_set or args.data_set>=len(mod.data):
            iprint("The selected data set does not exist - select one between 0 and %d"%(len(mod.data)-1))
            return
        iprint('Exporting data set %d into ASCII file %s'%(args.data_set, save_datafile))
        mod.data[args.data_set].save_file(save_datafile)


def set_numba_single():
    config_path = os.path.abspath(appdirs.user_data_dir('GenX3', 'ArturGlavic'))
    cache_dir = os.path.join(config_path, 'single_cpu_numba_cache')

    debug('Setting numba JIT compilation to single CPU')
    import numba
    numba.config.CACHE_DIR = cache_dir
    old_jit = numba.jit

    def jit(*args, **opts):
        opts['parallel'] = False
        return old_jit(*args, **opts)

    numba.jit = jit
    numba.GENX_OVERWRITE_SINGLE = True
    try:
        numba.set_num_threads(1)
    except AttributeError:
        pass


class InputThread(Thread):
    def __init__(self):
        Thread.__init__(self, daemon=True)
        self.stop_fit = False

    def run(self):
        # concole input has to be queried differently for non-UNIX systems
        if sys.platform.startswith('win'):
            self.run_windows()
        else:
            self.run_unix()

    def run_windows(self):
        import msvcrt, time
        while not self.stop_fit:
            if msvcrt.kbhit():
                debug('Key event received')
                key = msvcrt.getwch()
                if 'q' in key:
                    self.stop_fit = True
            time.sleep(0.1)

    def run_unix(self):
        import fcntl, selectors, time
        # set sys.stdin non-blocking
        orig_fl = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
        fcntl.fcntl(sys.stdin, fcntl.F_SETFL, orig_fl | os.O_NONBLOCK)

        m_selector = selectors.DefaultSelector()
        m_selector.register(sys.stdin, selectors.EVENT_READ)
        while True:
            (k, evt)=m_selector.select()[0]
            res = k.fileobj.read()
            if 'q' in res:
                self.stop_fit = True
                break
        m_selector.close()

def start_fitting(args, rank=0):
    """
    Function to start fitting from the command line.
    """
    import time
    from .model_control import ModelController
    from .core import config as io
    from .core.console import calc_errorbars, setup_console

    # Open the genx file
    io.config.load_default(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'profiles', 'default.profile'))
    if not args.bumps:
        from .diffev import DiffEv
        ctrl = ModelController(DiffEv())
        mod = ctrl.model
        opt: DiffEv = ctrl.optimizer
        set_optimiser_pars = set_diffev_pars
    else:
        from .bumps_optimizer import BumpsOptimizer
        ctrl = ModelController(BumpsOptimizer())
        mod = ctrl.model
        opt: BumpsOptimizer = ctrl.optimizer
        set_optimiser_pars = set_bumps_pars

    if rank==0:
        setup_console(ctrl, args.error, args.outfile, use_curses=(args.use_curses and not args.mpi))

    if (args.mpi or args.pr>0) and not args.disable_numba:
        try:
            set_numba_single()
        except ImportError:
            pass

    if rank==0:
        iprint('Loading model %s...'%args.infile)
    ctrl.load_file(args.infile)
    # has to be used in order to save everything....
    if args.esave:
        io.config.set('solver', 'save all evals', True)
    # Simulate, this will also compile the model script
    if rank==0:
        iprint('Simulating model...')
    mod.simulate()

    # Sets up the fitting ...
    set_optimiser_pars(opt, args)
    if rank==0:
        iprint('Setting up the optimizer...')
        iprint(opt)

    if args.outfile and rank==0:
        iprint('Saving the initial model to %s'%args.outfile)
        ctrl.save_file(args.outfile)

    # To start the fitting
    if rank==0:
        iprint('Fitting starting...')
        t1 = time.time()
    # print opt.use_mpi, opt.use_parallel_processing
    opt.start_fit(mod)
    if rank==0:
        inp = InputThread()
        inp.start()
    else:
        class Inp(): stop_fit = False
        inp = Inp()
    while opt.is_running():
        try:
            time.sleep(0.1)
            if inp.stop_fit:
                opt.stop = True
        except KeyboardInterrupt:
            iprint('KeyboardInterrupt, trying to stop fit.')
            opt.stop = True

    if rank==0:
        t2 = time.time()
        iprint('Fitting finished!')
        iprint('Time to fit: ', (t2-t1)/60., ' min')

    if rank==0:
        iprint('Updating the parameters')
        mod.parameters.set_value_pars(opt.best_vec)

    if args.outfile and rank==0:
        if args.error:
            iprint('Calculating errorbars')
            calc_errorbars(mod, opt)
        iprint('Saving the fit to %s'%args.outfile)
        # opt.set_use_mpi(False)
        ctrl.save_file(args.outfile)

    if rank==0:
        iprint('Fitting successfully completed')


def set_diffev_pars(optimiser, args):
    """
    Sets the optimiser parameters from args
    """
    if args.pr:
        optimiser.set_processes(args.pr)
        optimiser.set_use_parallel_processing(True)

    if args.mpi:
        optimiser.set_use_mpi(True)
    else:
        optimiser.set_use_mpi(False)

    if args.cs:
        optimiser.set_chunksize(args.cs)

    if args.mgen:
        optimiser.set_max_generations(args.mgen)
        optimiser.set_use_max_generations(True)

    if args.pops:
        optimiser.set_pop_size(args.pops)
        optimiser.set_use_pop_mult(False)

    if args.asi:
        optimiser.set_autosave_interval(args.asi)
        optimiser.set_use_autosave(True)

    if args.var>=0:
        optimiser.opt.min_parameter_spread = args.var

    if args.km>=0:
        optimiser.set_km(args.km)

    if args.kr>=0:
        optimiser.set_kr(args.kr)


def set_bumps_pars(optimiser, args):
    if args.pr:
        optimiser.opt.use_parallel_processing = True
        optimiser.opt.parallel_processes = args.pr

def compile_numba():
    try:
        # perform a compilation of numba functions with console feedback
        import numba

        real_jit = numba.jit

        class UpdateJit:
            update_counter = 1

            def __call__(self, *args, **opts):
                print(f'compiling numba functions {self.update_counter}/21')
                self.update_counter += 1
                return real_jit(*args, **opts)

        print('Starting to compile numba functions..')
        numba.jit = UpdateJit()
        from .models.lib import paratt_numba, neutron_numba, instrument_numba, offspec, surface_scattering
        numba.jit = real_jit
    except Exception as e:
        print('An exception occured when trying to compile the numba functions:')
        print(e)
        return 1
    return 0

def main():
    multiprocessing.freeze_support()
    multiprocessing.set_start_method('spawn')
    # Attempt to load mpi:
    try:
        from mpi4py import MPI
    except ImportError:
        rank = 0
        __mpi__ = False
    else:
        __mpi__ = bool(MPI.COMM_WORLD.Get_size()>1)
        rank = MPI.COMM_WORLD.Get_rank()

    parser = argparse.ArgumentParser(description="GenX %s, fits data to a model."%version.__version__,
                                     epilog="For support, manuals and bug reporting see http://genx.sf.net"
                                     )
    run_group = parser.add_mutually_exclusive_group()
    run_group.add_argument('-r', '--run', action='store_true', help='run GenX fit (no gui)')
    if __mpi__:
        run_group.add_argument('--mpi', action='store_true', help='run GenX fit with mpi (no gui)')
    run_group.add_argument('-g', '--gen', action='store_true', help='generate data.y with poisson noise added')
    run_group.add_argument('--pars', action='store_true', help='extract the parameters from the infile')
    run_group.add_argument('--mod', action='store_true', help='modify the GenX file')
    opt_group = parser.add_argument_group('optimization arguments')
    opt_group.add_argument('--pr', type=int, default=0, help='Number of processes used in parallel fitting.')
    opt_group.add_argument('--cs', type=int, default=0, help='Chunk size used for parallel processing.')
    opt_group.add_argument('--mgen', type=int, default=0, help='Maximum number of generations that is used in a fit')
    opt_group.add_argument('--pops', type=int, default=0, help='Population size - number of individuals.')
    opt_group.add_argument('--asi', type=int, default=0, help='Auto save interval (generations).')
    opt_group.add_argument('--km', type=float, default=-1, help='Mutation constant (float 0 < km < 1)')
    opt_group.add_argument('--kr', type=float, default=-1, help='Cross over constant (float 0 < kr < 1)')
    opt_group.add_argument('-s', '--esave', action='store_true', help='Force save evals to gx file.')
    opt_group.add_argument('-e', '--error', action='store_true', help='Calculate/export error bars.')
    opt_group.add_argument('--var', type=float, default=-1,
                           help='Minimum relative parameter variation to stop the fit (%%)')
    opt_group.add_argument('--bumps', action='store_true',
                           help='Use Bumps DREAM optimizer instead of GenX Differential Evolution')
    data_group = parser.add_argument_group('data arguments')
    data_group.add_argument('-d', dest='data_set', type=int, default=0,
                            help='Active data set to act upon. Index starting at 0.')
    data_group.add_argument('--load', dest='datafile',
                            help='Load file into active data set. Index starting at 0.')
    data_group.add_argument('--export', dest='save_datafile',
                            help='Save active data set to file. Index starting at 0.')
    data_group = parser.add_argument_group('startup options')
    data_group.add_argument('-l', '--logfile', dest='logfile', default=None, type=str,
                            help='Output debug information to logfile.')
    data_group.add_argument('--debug', dest='debug', default=False, action="store_true",
                            help='Show additional debug information on console/logfile')
    data_group.add_argument('--dpi-scale', dest='dpi_overwrite', default=None, type=float,
                            help='Overwrite the detection of screen dpi scaling factor (=72/dpi)')
    data_group.add_argument('--no-curses', dest='use_curses', default=True, action="store_false",
                            help='Disable Curses interactive console interface for command line '
                                 'fitting on UNIX systems.')
    data_group.add_argument('--disable-nb', dest='disable_numba', default=False, action="store_true",
                            help='Disable the use of numba JIT compiler')
    data_group.add_argument('--nb1', dest='numba_single', default=False, action="store_true",
                            help='Compile numba JIT functions without parallel computing support (use one core only). '
                                 'Caching in this case is done in a different user directory.')
    data_group.add_argument('--compile-nb', dest='compile_nb', default=False, action="store_true",
                            help='Perform a first-/recompilation of the numba modules and exit')

    parser.add_argument('infile', nargs='?', default='',
                        help='The .gx or .hgx file to load or .ort file to use as basis for model')
    parser.add_argument('outfile', nargs='?', default='', help='The .gx  or hgx file to save into')

    args = parser.parse_args()
    if not __mpi__:
        args.mpi = False

    if args.compile_nb:
        sys.exit(compile_numba())

    if args.run or args.mpi or args.pars or args.mod:
        # make sure at least info-messages are shown (default is warning)
        custom_logging.CONSOLE_LEVEL = min(logging.INFO, custom_logging.CONSOLE_LEVEL)
    if rank>0:
        custom_logging.CONSOLE_LEVEL = logging.WARNING
    setup_system()

    if args.logfile:
        if __mpi__:
            lfbase, lfend = args.logfile.rsplit('.', 1)
            activate_logging(f'{lfbase}_{rank:02}.{lfend}')
        else:
            activate_logging(args.logfile)
    debug("Arguments from parser: %s"%args)
    if not args.outfile:
        args.outfile = args.infile
    args.outfile = os.path.abspath(args.outfile)
    if args.infile:
        args.infile = os.path.abspath(args.infile)

    if args.disable_numba:
        debug('disable numba')
        # set numba flag
        from genx.models import lib as modellib
        modellib.USE_NUMBA = False
    elif args.numba_single:
        set_numba_single()

    if args.run:
        start_fitting(args)
    elif args.mpi:
        start_fitting(args, rank)
    elif args.gen:
        create_simulated_data(args)
    elif args.pars:
        extract_parameters(args)
    elif args.mod:
        modify_file(args)
    elif not args.run and not args.mpi:
        # Check if the application has been frozen
        if hasattr(sys, "frozen") and True:
            # Redirect all the output to log files
            log_file_path = appdirs.user_log_dir('GenX3', 'ArturGlavic')
            # Create dir if not found
            if not os.path.exists(log_file_path):
                os.makedirs(log_file_path)
            # print log_file_path
            # log_file_path = genx_gui._path + 'app_data/'
            sys.stdout = open(log_file_path+'/genx.log', 'w')
            sys.stderr = open(log_file_path+'/genx.log', 'w')
        start_interactive(args)


if __name__=="__main__":
    main()



"""
DWBA extension for offspecular from rough surfaces.

Rebuild of old C++ extension module.
"""
from numpy import *
from scipy import integrate
from scipy.special import factorial
from .elfield import *

import numba

@numba.jit(numba.float64[:](numba.float64[:], numba.complex128[:,:,:], numba.complex128[:,:,:], numba.float64,
                            numba.float64, numba.float64[:], numba.float64[:], numba.complex128[:], numba.float64[:],
                            numba.complex128[:], numba.float64[:], numba.float64[:], numba.float64),
            nopython=True, parallel=True, cache=True)
def dwba_interdiff_sum(qx, G, q, eta,
                       h, sigma, sigmaid, sqn, z,
                       table, q_min, fn, eta_z):
    Layers=G.shape[1]
    Points=qx.shape[0]

    s=0.0j
    I=zeros(qx.shape, dtype=float64)

    for m in numba.prange(Points):
        Itemp=0.0j
        for i in range(Layers):
            for j in range(Layers):
                for k in range(4):
                    for l in range(4):
                        nind=(k, i, m) # DWBA component, Layer, q-point for incident
                        npind=(l, j, m) # DWBA component, Layer, q-point for outgoing
                        for p in range(fn.shape[0]):
                            xnew=abs(qx[m]*eta/pow(p+1.0, 1.0/2.0/h))
                            lower=int((xnew-q_min[0])/(q_min[1]-q_min[0]))
                            s+=2.0*pow(q[nind]*conj(q[npind])*sigma[i]*sigma[j]*exp(-abs(z[i]-z[j])/eta_z), p+1.0)\
                               /fn[p]*eta/pow(p+1.0, 1.0/2.0/h)*\
                               ((table[lower+1]-table[lower])/(q_min[lower+1]-q_min[lower])*
                                (xnew-q_min[lower])+table[lower]);
                        Itemp+=(sqn[i]-sqn[i+1])*conj(sqn[j]-sqn[j+1])*\
                               G[nind]*exp(-0.5*pow(sigmaid[i]*q[nind], 2))*\
                               conj(G[npind]*exp(-0.5*pow(sigmaid[j]*q[npind], 2)))*\
                               exp(-0.5*(pow(q[nind]*sigma[i], 2)+pow(conj(q[npind])*sigma[j], 2)))\
                               /q[nind]/conj(q[npind])*s
                        s = 0
        I[m]=Itemp.real
    return I

@numba.jit(numba.float64[:](numba.float64[:], numba.complex128[:,:,:], numba.complex128[:,:,:], numba.float64,
                            numba.float64, numba.float64[:], numba.complex128[:], numba.float64[:],
                            numba.complex128[:], numba.float64[:], numba.float64[:], numba.float64),
            nopython=True, parallel=True, cache=True)
def dwba_sum(qx, G, q, eta,
             h, sigma, sqn, z,
             table, q_min, fn, eta_z):
    Layers=G.shape[1]
    Points=qx.shape[0]
    s=0.0j
    I=zeros(qx.shape, dtype=float64)

    for m in numba.prange(Points):
        Itemp=0.0j
        for i in range(Layers):
            for j in range(Layers):
                for k in range(4):
                    for l in range(4):
                        nind=(k, i, m) # DWBA component, Layer, q-point for incident
                        npind=(l, j, m) # DWBA component, Layer, q-point for outgoing
                        for p in range(fn.shape[0]):
                            xnew = fabs(qx[m]*eta/pow(p+1.0,1.0/2.0/h))
                            lower=int((xnew-q_min[0])/(q_min[1]-q_min[0]))
                            s+=2.0*pow(q[nind]*conj(q[npind])*sigma[i]*sigma[j]*exp(-abs(z[i]-z[j])/eta_z),p+1.0)\
                               /fn[p]*eta/pow(p+1.0,1.0/2.0/h)*\
                               ((table[lower+1]-table[lower])/(q_min[lower+1]-q_min[lower])*
                                (xnew-q_min[lower])+table[lower])
                        Itemp+=(sqn[i]-sqn[i+1])*conj(sqn[j]-sqn[j+1])*\
                               G[nind]*conj(G[npind])*\
                               exp(-0.5*(pow(q[nind]*sigma[i],2.0)+pow(conj(q[npind])*sigma[j],2.0)))\
                               /q[nind]/conj(q[npind])*s
                        s=0.0j
        I[m]=Itemp.real
    return I

@numba.jit(numba.float64[:](numba.float64[:], numba.float64[:], numba.float64,
                            numba.float64, numba.float64[:], numba.complex128[:], numba.float64[:],
                            numba.complex128[:], numba.float64[:], numba.float64[:], numba.float64),
            nopython=True, parallel=True, cache=True)
def born_ext(qx, qz, eta,
             h, sigma, sqn, z,
             table, q_min, fn, eta_z):
    Layers=sigma.shape[0]
    Points=qx.shape[0]
    s=0.0j
    I=zeros(qx.shape, dtype=float64)

    for m in numba.prange(Points):
        Itemp=0.0j
        for i in range(Layers):
            for j in range(Layers):
                for p in range(fn.shape[0]):
                    xnew=qx[m]*eta/pow(p+1.0, 1.0/2.0/h)
                    lower=int((xnew-q_min[0])/(q_min[1]-q_min[0]))
                    s+=2.0*pow(qz[m]*qz[m]*sigma[i]*sigma[j]*exp(-abs(z[i]-z[j])/eta_z), p+1.0)/fn[p]*eta/pow(
                        p+1.0, 1.0/2.0/h)*((table[lower+1]-table[lower])/(q_min[lower+1]-q_min[lower])*(
                                xnew-q_min[lower])+table[lower])

                Itemp+=(sqn[i]-sqn[i+1])*conj(sqn[j]-sqn[j+1])*exp(
                    -0.5*(pow(qz[m]*sigma[i], 2.0)+pow(qz[m]*sigma[j], 2.0)))/qz[m]/qz[m]*s;
                s=0.0j
        I[m]=Itemp.real
    return I


def vec_realsymint(F, omega, eta, h, eta_z, qz_n, qz_np, sigma_n, sigma_np, max_n):
    I=2*sum(
        [((qz_n*conj(qz_np)*(sigma_n*sigma_np))**n)/factorial(n)*eta/n**(1.0/2.0/h)*F(omega*eta/(n**(1.0/2.0/h))) for n
         in arange(1.0, max_n+1.0)])
    I=array(I)
    return I

def make_F(q, h):
    def f(x):
        return exp(-abs(x)**(2*h))

    q=q*(abs(q)>0.003)+0.003*(abs(q)<0.003)
    I=[integrate.quad(f, 0, inf, weight='cos', wvar=x)[0] for x in q]
    I=array(I)
    return I

def DWBA_Interdiff(qx, qz, lamda, n, z, sigma, sigmaid, eta, h, eta_z, d=[0],
                   taylor_n=1):
    # sigmaid-interdiffused part
    # sigma-roughness
    k=2*pi/lamda
    qx=array(qx, dtype=float64)
    # Calculating electrical fields
    omega=arctan(qx/qz)  # Not the real omega given by Fewster
    omegap=arcsin(sqrt(qx**2+qz**2)/2/k)
    kxi=k*cos(omega+omegap)
    kxf=k*cos(omegap-omega)

    # Calculate the electric fields
    (T, R, kz)=AmpElfield_q(k, r_[kxi, kxf], lamda, n[::-1], r_[0, d[1:][::-1]])

    lenq=len(qx)
    Ri=R[:, :lenq]
    Rf=R[:, lenq:]
    Ti=T[:, :lenq]
    Tf=T[:, lenq:]
    ki=kz[:, :lenq]
    kf=kz[:, lenq:]
    G=array([Ti*Tf, Ti*Rf, Ri*Tf, Ri*Rf], dtype=complex128)
    q=array([ki+kf, ki-kf, -ki+kf, -ki-kf], dtype=complex128)

    # Setting up for the Fourier integral as given by Pape et.al.
    maxn=taylor_n
    # New dqmin tested 20051201
    dqmin=min(abs(qx[:-1]-qx[1:]))*eta/(maxn**(1.0/2.0/h))
    if abs(dqmin)<1e-12:
        dqmin=1e-3
    # New q_min tested 20051201
    q_min=arange(min(abs(qx))*eta/(maxn**(1.0/2.0/h))-dqmin, max(abs(qx))*eta+2*dqmin, dqmin, dtype=float64)
    table=array(make_F(q_min, h), dtype=complex128)

    fn=factorial(arange(1, maxn+1))
    sqn=n**2
    sqn=array(sqn, dtype=complex128)
    sigma=array(sigma, dtype=float64)
    sigmaid=array(sigmaid, dtype=float64)
    s=dwba_interdiff_sum(qx, G, q, eta, h, sigma, sigmaid, sqn, z, table, q_min, fn, eta_z)

    return (s, omega+omegap, omegap-omega)

def DWBA(qx, qz, lamda, n, z, sigma, eta, h, eta_z, d=[0], taylor_n=1):
    k=2*pi/lamda
    qx=array(qx, dtype=float64)
    sqn=n**2
    sqn=array(sqn, dtype=complex)
    # Calculating electrical fields
    omega=arctan(qx/qz)  # Not the real omega given by Fewster
    omegap=arcsin(sqrt(qx**2+qz**2)/2/k)
    kxi=k*cos(omega+omegap)
    kxf=k*cos(omegap-omega)
    (T, R, kz)=AmpElfield_q(k, r_[kxi, kxf], lamda, n[::-1], r_[0, d[1:][::-1]])

    lenq=len(qx)
    Ri=R[:, :lenq]
    Rf=R[:, lenq:]
    Ti=T[:, :lenq]
    Tf=T[:, lenq:]
    ki=kz[:, :lenq]
    kf=kz[:, lenq:]
    G=array([Ti*Tf, Ti*Rf, Ri*Tf, Ri*Rf], dtype=complex)
    q=array([ki+kf, ki-kf, -ki+kf, -ki-kf], dtype=complex)

    # Setting up for the Fourier integral as given by Pape et.al.
    maxn=taylor_n
    dqmin=(qx[1]-qx[0])*eta/(maxn**(1.0/2.0/h))
    if abs(dqmin)<1e-12:
        dqmin=1e-3
    q_min=arange(qx[0]*eta/(maxn**(1.0/2.0/h)), qx[-1]*eta+2*dqmin, dqmin, dtype=float64)

    table=array(make_F(q_min, h), dtype=complex)
    fn=factorial(arange(1, maxn+1))
    s=dwba_sum(qx, G, q, eta, h, sigma, sqn, z, table, q_min, fn, eta_z)
    return (s, omega+omegap, omegap-omega)

def Born(qx, qz, lamda, n, z, sigma, eta, h, eta_z, d=[0], taylor_n=1):
    k=2*pi/lamda
    qx=array(qx, dtype=float64)
    sqn=n**2
    sqn=array(sqn, dtype=complex)
    # print 'Setup Complete'
    # Calculating electrical fields
    omega=arctan(qx/qz)  # Not the real omega given by Fewster
    omegap=arcsin(sqrt(qx**2+qz**2)/2/k)
    # Setting up for the Fourier integral as given by Pape et.al.
    maxn=taylor_n
    dqmin=(qx[1]-qx[0])*eta/(maxn**(1.0/2.0/h))
    q_min=arange(qx[0]*eta/(maxn**(1.0/2.0/h)), qx[-1]*eta+2*dqmin, dqmin, dtype=float64)

    table=make_F(q_min, h)
    table=array(table, dtype=complex128)

    fn=factorial(arange(1, maxn+1))
    s=born_ext(qx, qz, eta, h, sigma, sqn, z, table, q_min, fn, eta_z)

    return (s, omega+omegap, omegap-omega)

if __name__=='__main__':
    n=array([1]+[1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j]*10+[1-7.57e-6+1.73e-7j])
    # n=array([1,1-2.25e-5+2.9e-6j,1-7.5e-6+1.75e-7j])
    # n=array([1,1-2.25e-5,1-7.5e-6])
    d=array([0.0]+[80.0, 20.0]*10)
    # d=array([0.0,1000.0])
    z=-cumsum(d)
    sigma=array([5.0]+[5.0, 5.0]*10)
    # sigma=array([3.0,3.0])

    qx=arange(.1e-4, 0.005, 0.00002)
    qz=0.2*ones(qx.shape)

    # qx=arange(1e-5,3e-5,1e-5)
    eta=200.0
    eta_z=1.0
    h=1.0

    (I1, ain, aout)=Born(qx, qz, 1.540562, n, z, sigma, eta, h, eta_z)
    (I2, ain, aout)=DWBA(qx, qz, 1.540562, n, z, sigma, eta, h, eta_z, d)
    (I3, ain, aout)=DWBA_Interdiff(qx, qz, 1.540562, n, z, sigma, 0.3*sigma, eta, h, eta_z, d)

    from matplotlib import pyplot
    pyplot.plot(qx, real(I1), qx, real(I2), qx, real(I3))
    pyplot.show()



'''
========================================
:mod:`d17_cosmos` D17 cosmos data loader
========================================

Loads .out files from the cosmos program that is used to process time of flight
data from the D17 iunstrument at the ILL. 

The default columns are the following:

* First column q values
* Second column Intensitiy values
* Third values The unceratinty in the Intensities
* Fourth column q-resolution

The data loader skips the first 36 lines of the file which is assumed to be the header.
The other settings are just as in the default data loader.

The resolution is stored as the member variable res. Can be accessed, for data set 0, 
data[0].x
'''

import numpy as np

from ..data_loader_framework import Template
from ..utils import ShowWarningDialog

try:
    import wx
    from wx.lib.masked import NumCtrl
except ImportError:
    class void():
        pass

    wx=void()
    wx.Dialog=void

class Plugin(Template):
    wildcard='*.mft'

    def __init__(self, parent=None):
        Template.__init__(self, parent)
        self.q_col=0
        self.I_col=1
        self.eI_col=2
        self.res_col=3
        self.comment='#'
        self.skip_rows=23
        self.delimiter=None

    def LoadData(self, dataset, filename, data_id=0):
        '''LoadData(self, data_item_number, filename) --> none
        
        Loads the data from filename into the data_item_number.
        '''
        try:
            with open(filename, 'r', encoding='utf8') as fh:
                header=fh.readlines()[:self.skip_rows]
            load_array=np.loadtxt(filename, delimiter=self.delimiter, encoding='utf8',
                                  comments=self.comment, skiprows=self.skip_rows)
        except Exception as e:
            ShowWarningDialog(self.parent, 'Could not load the file: '+ \
                              filename+' \nPlease check the format.\n\n numpy.loadtxt' \
                              +' gave the following error:\n'+str(e))
        else:
            # For the freak case of only one data point
            if len(load_array.shape)<2:
                load_array=np.array([load_array])
            # Check so we have enough columns
            if load_array.shape[1]-1<max(self.q_col, self.res_col,
                                         self.I_col, self.eI_col):
                ShowWarningDialog(self.parent, 'The data file does not contain' \
                                  +'enough number of columns. It has '+str(load_array.shape[1]) \
                                  +' columns. Rember that the column index start at zero!')
                # Okay now we have showed a dialog lets bail out ...
                return
            # The data is set by the default Template.__init__ function, neat hu
            # Note that the loaded data goes into *_raw so that they are not
            # changed by the transforms
            header_info={}
            for hline in header[:-2]:
                try:
                    key, value=map(str.strip, hline.split(':', 1))
                except:
                    pass
                else:
                    header_info[key]=value

            # print load_array
            dataset.x_raw=load_array[:, self.q_col]
            dataset.y_raw=load_array[:, self.I_col]
            dataset.error_raw=load_array[:, self.eI_col]
            dataset.set_extra_data('res', load_array[:, self.res_col], 'res/(2.*sqrt(2.*log(2)))')
            dataset.res=load_array[:, self.res_col]*1.0
            # Run the commands on the data - this also sets the x,y, error memebers
            # of that data item.
            dataset.run_command()

            # insert metadata into ORSO compatible fields
            dataset.meta['data_source']['facility']='ILL'
            dataset.meta['data_source']['experiment']['instrument']='D17'
            dataset.meta['data_source']['experiment']['probe']='neutron'
            if 'Title' in header_info:
                dataset.meta['data_source']['title']=header_info['Title']
            if 'Start date + time' in header_info:
                dataset.meta['data_source']['experimentDate']=header_info['Start date + time']
            dataset.meta['data_source']['file_header']=header_info

    def SettingsDialog(self):
        '''SettingsDialog(self) --> None
        
        This function should - if necessary implement a dialog box
        that allows the user set import settings for example.
        '''
        col_values={'I': self.I_col, 'q': self.q_col, 'Resolution': self.res_col,
                    'I error': self.eI_col}
        misc_values={'Comment': str(self.comment), 'Skip rows': self.skip_rows,
                     'Delimiter': str(self.delimiter)}
        dlg=SettingsDialog(self.parent, col_values, misc_values)
        if dlg.ShowModal()==wx.ID_OK:
            col_values=dlg.GetColumnValues()
            misc_values=dlg.GetMiscValues()
            self.q_col=col_values['q']
            self.res_col=col_values['Resolution']
            self.I_col=col_values['I']
            self.eI_col=col_values['I error']
            self.comment=misc_values['Comment']
            self.skip_rows=misc_values['Skip rows']
            self.delimiter=misc_values['Delimiter']
            self.SetStatusText('Changed import settings')
        else:
            self.SetStatusText('No change to import settings')
        dlg.Destroy()

class SettingsDialog(wx.Dialog):

    def __init__(self, parent, col_values, misc_values):
        wx.Dialog.__init__(self, parent, -1, 'Data loader settings')

        box_sizer=wx.BoxSizer(wx.HORIZONTAL)

        # Make the box for putting in the columns
        col_box=wx.StaticBox(self, -1, "Columns")
        col_box_sizer=wx.StaticBoxSizer(col_box, wx.VERTICAL)

        # col_values = {'y': 1,'x': 0,'y error': 1}
        col_grid=wx.GridBagSizer(len(col_values), 2)
        self.col_controls=col_values.copy()
        keys=list(col_values.keys())
        keys.sort()
        for i, name in enumerate(keys):
            text=wx.StaticText(self, -1, name+': ')
            control=wx.SpinCtrl(self)
            control.SetRange(0, 100)
            control.SetValue(col_values[name])
            col_grid.Add(text, (i, 0),
                         flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL,
                         border=5)
            col_grid.Add(control, (i, 1),
                         flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL,
                         border=5)
            self.col_controls[name]=control

        col_box_sizer.Add(col_grid, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        box_sizer.Add(col_box_sizer, 0, wx.ALL | wx.EXPAND, 5)

        col_box=wx.StaticBox(self, -1, "Misc")
        col_box_sizer=wx.StaticBoxSizer(col_box, wx.VERTICAL)

        # Lets add another box for comments and rows to skip
        # misc_values = {'Comment': '#', 'Skip rows': 0,'Delimiter': 'None'}
        col_grid=wx.GridBagSizer(len(misc_values), 2)
        self.misc_controls=misc_values.copy()
        keys=list(misc_values.keys())
        keys.sort()
        for i, name in enumerate(keys):
            text=wx.StaticText(self, -1, name+': ')
            if type(misc_values[name])==type(1):
                control=wx.SpinCtrl(self)
                control.SetRange(0, 100)
                control.SetValue(misc_values[name])
            else:
                control=wx.TextCtrl(self, value=misc_values[name],
                                    style=wx.EXPAND)
            col_grid.Add(text, (i, 0),
                         flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL,
                         border=5)
            col_grid.Add(control, (i, 1),
                         flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL,
                         border=5)
            self.misc_controls[name]=control

        col_box_sizer.Add(col_grid, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        box_sizer.Add(col_box_sizer, 0, wx.ALL | wx.EXPAND, 5)

        button_sizer=wx.StdDialogButtonSizer()
        okay_button=wx.Button(self, wx.ID_OK)
        okay_button.SetDefault()
        button_sizer.AddButton(okay_button)
        button_sizer.AddButton(wx.Button(self, wx.ID_CANCEL))
        button_sizer.Realize()

        sizer=wx.BoxSizer(wx.VERTICAL)
        sizer.Add(box_sizer, 1, wx.GROW, 20)
        line=wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW, 30)

        sizer.Add(button_sizer, 0,
                  flag=wx.ALIGN_RIGHT, border=20)
        self.SetSizer(sizer)

        sizer.Fit(self)
        self.Layout()

    def GetColumnValues(self):
        values={}
        for key in self.col_controls:
            values[key]=self.col_controls[key].GetValue()
        return values

    def GetMiscValues(self):
        values={}
        for key in self.misc_controls:
            val=self.misc_controls[key].GetValue()
            if type(val)==type('') or type(val)==type(''):
                if val.lower()=='none':
                    val=None
            values[key]=val
        return values



'''
Script to download the FFAST library and convert it to nff files
'''

__author__='Matts Bjorck'

import urllib.request, urllib.parse, urllib.error, numpy, os, time

def get_html_code(Z):
    adress='http://physics.nist.gov/cgi-bin/ffast/ffast.pl?Z=%d&Formula=&gtype=4&lower=&upper=&density=&frames=no'%Z
    f=urllib.request.urlopen(adress)
    return f.readlines()

def find_line(lines, string_to_find):
    for i, line in enumerate(lines):
        if line.find(string_to_find)!=-1:
            return line, i
    return None, None

def parse_page(lines):
    # Find the element
    line, i=find_line(lines, '(Z =')
    element=line.split('<b>')[1].split('&#160;')[0].strip()
    print('Element: ', element)

    # Find the Relativistic correction estimate
    line, i=find_line(lines, 'Relativistic correction estimate')
    rel_corr=float(line.split('=')[1].split(',')[1].split('<i>e</i>')[0].strip())
    print('Relativistic correction estimate 3/5CL: ', rel_corr)

    # Find the Nuclear Thomson correction
    line, i=find_line(lines, 'Nuclear Thomson correction')
    nuc_corr=float(line.split('=')[1].split('<i>e</i>')[0].strip())
    print('Nuclear Thomson correction: ', nuc_corr)

    # Find the data
    line, i=find_line(lines, '<b>Form Factors,')
    data_lines=lines[i+3:-3]
    data=numpy.fromstring(' '.join(data_lines), sep=' ').reshape((len(data_lines), 8))
    table=numpy.c_[data[:, 0]*1e3, data[:, 1]+rel_corr+nuc_corr, data[:, 2]]
    return table, element

def create_nff(Z, path):
    html_code=get_html_code(Z)
    table, element=parse_page(html_code)
    with open(os.path.join(path, '%s.nff'%element.lower()), 'w') as f:
        f.write('E (eV)\tf1 e/atom\tf2 e/atom\n')
        numpy.savetxt(f, table)

def create_readme(path):
    readme_str=(
            '''README
    These tables were generated %s.
    The following files were created by the download_nist.py script located in the GenX distribution. The data
    were generated from the NIST X-ray Form Factor, Attenuation and Scattering Tables (FFAST) located at:
    http://www.nist.gov/pml/data/ffast/index.cfm.
    
    Each element's scattering factors were evaluated according to eq. 4 in
    C. T. Chandler J. Phys. Chem. Ref. Data 29(4) 597-1048 (2000). Thus the scattering factors were calculated
    according to
        f1 = f_1 + f_rel(3/5 CL) + f_NT,
        f2 = f_2.
    
    According to the homepage, see above, the tables should be referenced according to:
        Chantler, C.T., Olsen, K., Dragoset, R.A., Chang, J., Kishore, A.R., Kotochigova, S.A., and Zucker, D.S. (2005),
        X-Ray Form Factor, Attenuation and Scattering Tables (version 2.1).
        [Online] Available: http://physics.nist.gov/ffast [%s].
        National Institute of Standards and Technology, Gaithersburg, MD.
        Originally published as Chantler, C.T., J. Phys. Chem. Ref. Data 29(4), 597-1048 (2000);
        and Chantler, C.T., J. Phys. Chem. Ref. Data 24, 71-643 (1995).
            '''%(time.ctime(), time.ctime())
    )
    with open(os.path.join(path, 'README.txt'), 'w') as f:
        f.write(readme_str)

if __name__=='__main__':
    lib_path='f1f2_nist'
    for z in range(1, 93):
        create_nff(z, lib_path)
    create_readme(lib_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
.. module:: parameters
   :synopsis: Classes to work with parameters and expressions thereof.

.. moduleauthor:: Matts Björck <matts.bjorck@gmail.com>

"""
import numpy as np

from collections.abc import MutableSequence
from genx.core.custom_logging import iprint

class Parameter(object):
    """Base class for all Parameters"""

    def __init__(self, help="", unit=""):
        self.help=help
        self.unit=unit

    def _get_value(self, **kwargs):
        raise NotImplementedError("Callback parameter method not implemented")

    def __call__(self, **kwargs):
        return self._get_value(**kwargs)

    def validate(self, value):
        """ Function to test that value is valid. Should raise a ValueError otherwise.
        """
        raise NotImplementedError("Validate method not implemented")

    def couple_parameter(self, parameter):
        """Function to call when a parameter is coupled (replaced) by another.
        This is used in HasParameters classes.
        """
        raise AttributeError('This parameter can not be set to another parameter')

    def has_coupled_parameter(self):
        """Returns True if the parameter is coupled to another parameter."""
        return False

def is_parameter(obj):
    """Returns true if object is a subclass of the class Parameter"""
    return isinstance(obj, Parameter)

def is_hasparameter(obj):
    """Returns true if object is a subclass of the class HasParameter"""
    return isinstance(obj, HasParameters)

class ArithmeticParameter(Parameter):
    """ A parameter that supports arithmetic calculations"""

    def _check_obj(self, other):
        """ Checks the object other so that it fulfills the demands for arithmetic operations.
        """
        supported_types=[int, float, int, complex, np.float64, np.float32]
        if is_parameter(other):
            pass
        elif not type(other) in supported_types:
            raise TypeError("%s is not supported for arithmetic operations "%(repr(type(other)))+
                            "of a Parameter. It has to be int, float, long or complex")

    def _new_object_from_func(self, value_func, help=""):
        """Create a new object from function"""
        new=ArithmeticParameter(help=help)
        new._get_value=value_func
        return new

    def __mul__(self, other):
        self._check_obj(other)
        if is_parameter(other):
            def new_func(**kwargs):
                return self(**kwargs)*other(**kwargs)
        else:
            def new_func(**kwargs):
                return self(**kwargs)*other
        return self._new_object_from_func(new_func)

    def __rmul__(self, other):
        self._check_obj(other)
        if is_parameter(other):
            def new_func(**kwargs):
                return other(**kwargs)*self(**kwargs)
        else:
            def new_func(**kwargs):
                return other*self(**kwargs)
        return self._new_object_from_func(new_func)

    def __add__(self, other):
        self._check_obj(other)
        if is_parameter(other):
            def new_func(**kwargs):
                return self(**kwargs)+other(**kwargs)
        else:
            def new_func(**kwargs):
                return self(**kwargs)+other
        return self._new_object_from_func(new_func)

    def __radd__(self, other):
        self._check_obj(other)
        if is_parameter(other):
            def new_func(**kwargs):
                return other(**kwargs)+self(**kwargs)
        else:
            def new_func(**kwargs):
                return other+self(**kwargs)
        return self._new_object_from_func(new_func)

    def __sub__(self, other):
        self._check_obj(other)
        if is_parameter(other):
            def new_func(**kwargs):
                return self(**kwargs)-other(**kwargs)
        else:
            def new_func(**kwargs):
                return self(**kwargs)-other
        return self._new_object_from_func(new_func)

    def __rsub__(self, other):
        self._check_obj(other)
        if is_parameter(other):
            def new_func(**kwargs):
                return other(**kwargs)-self(**kwargs)
        else:
            def new_func(**kwargs):
                return other-self(**kwargs)
        return self._new_object_from_func(new_func)

    def __div__(self, other):
        self._check_obj(other)
        if is_parameter(other):
            def new_func(**kwargs):
                return self(**kwargs)/other(**kwargs)
        else:
            def new_func(**kwargs):
                return self(**kwargs)/other
        return self._new_object_from_func(new_func)

    def __rdiv__(self, other):
        self._check_obj(other)
        if is_parameter(other):
            def new_func(**kwargs):
                return other(**kwargs)/self(**kwargs)
        else:
            def new_func(*args, **kwargs):
                return other/self(*args, **kwargs)
        return self._new_object_from_func(new_func)

    def __neg__(self):
        def new_func(**kwargs):
            return -self(**kwargs)

        return self._new_object_from_func(new_func)

    def __pos__(self):
        def new_func(**kwargs):
            return self(**kwargs)

        return self._new_object_from_func(new_func)

    def __pow__(self, other):
        self._check_obj(other)
        if is_parameter(other):
            def new_func(**kwargs):
                return self(**kwargs)**other(**kwargs)
        else:
            def new_func(**kwargs):
                return self(**kwargs)**other
        return self._new_object_from_func(new_func)

    def __rpow__(self, other):
        self._check_obj(other)
        if is_parameter(other):
            def new_func(**kwargs):
                return other(**kwargs)**self(**kwargs)
        else:
            def new_func(**kwargs):
                return other**self(**kwargs)
        return self._new_object_from_func(new_func)

class NumericParameter(ArithmeticParameter):
    """A parameter holding a numeric value"""

    def __init__(self, value, help="", unit=""):
        ArithmeticParameter.__init__(self, help, unit=unit)
        self.value=value
        self._coupled_parameter=None

    def validate(self, value):
        raise NotImplementedError("Validation of value not implemented")

    def __setattr__(self, key, value):
        if key=='value':
            object.__setattr__(self, 'value', self.validate(value))
            self._coupled_parameter=None
        else:
            object.__setattr__(self, key, value)

    def _get_value(self, **kwargs):
        """Callback for getting the value of the parameter"""
        if not self.has_coupled_parameter():
            return self.value
        else:
            return self.validate(self._coupled_parameter(**kwargs))

    def couple_parameter(self, parameter):
        """Couple a parameter to this parameter, replacing the _get_value output with its output"""
        object.__setattr__(self, '_coupled_parameter', parameter)

    def has_coupled_parameter(self):
        """Returns True if the parameter is coupled to another parameter."""
        return self._coupled_parameter is not None

class Enum(Parameter):
    """An enumeration object"""

    def __init__(self, allowed_values=None, help=""):
        Parameter.__init__(self, help=help)
        if allowed_values is None:
            allowed_values=[]
        self.allowed_values=allowed_values
        self.value=self.allowed_values[0]

    def __setattr__(self, key, value):
        if key=='value':
            object.__setattr__(self, 'value', self.validate(value))
        else:
            object.__setattr__(self, key, value)

    def validate(self, value):
        """Validates the value. Returns a proper value else raises an ValueError"""
        if value in self.allowed_values:
            return value
        elif isinstance(value, int):
            if len(self.allowed_values)>value>0:
                return self.allowed_values[0]
        raise ValueError("The value to an enum must be either an integer"
                         " between 0 and %d or one of %s"%(len(self.allowed_values), str(self.allowed_values)))

    def _get_value(self, **kwargs):
        return self.value

class List(MutableSequence):
    """A List object with type checking"""
    protected=True

    def __init__(self, allowed_type=None, value=None, help=""):
        if value is None:
            value=[]
        self.help=help
        self.allowed_type=allowed_type
        self.data=list()
        for item in value:
            self._check(item)
            self.data.append(item)

    def _check(self, value):
        if self.allowed_type is not None and not isinstance(value, self.allowed_type):
            raise ValueError("List object can only consist of type %s, trying to set a type: %s"%
                             (self.allowed_type, type(value)))

    def validate(self, value):
        self._check(value)
        return value

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key, value):
        # print "Setting ", key, " to ", value
        if isinstance(key, slice):
            for item in value:
                self._check(item)
        else:
            self.validate(value)
        self.data.__setitem__(key, value)

    def __delitem__(self, key):
        self.data.__delitem__(key)

    def __len__(self):
        return len(self.data)

    def insert(self, index, value):
        self._check(value)
        self.data.insert(index, value)

class HasParameters(object):
    """Mixin class to handle the setting of parameters in a correct manner.

    Attributes that are parameter can not be set to an incompatible type. This is done by overloading the __setattr__
    method.

    Also if an attribute implements a boolean attribute protected which is True then this attribute can not be set.
    """
    validation_kwargs={}

    def __init__(self, **kwargs):
        """The given allowed keyword arguments allow to initialise parameters"""
        for name in kwargs:
            if hasattr(self, name):
                self.__setattr__(name, kwargs[name])
            else:
                raise TypeError("Creation of %s got an unexpected keyword argument %s"%
                                (self.__class__.__name__, name))

    def __setattr__(self, name, value):
        try:
            attr=self.__getattribute__(name)
        except AttributeError:
            # Attribute does not exist
            object.__setattr__(self, name, value)
        else:
            if is_parameter(attr) and not self._is_protected(attr):
                if is_parameter(value) or is_hasparameter(value):
                    # Trying to set a Parameter to a Parameter...
                    # check so that the type is correct:
                    try:
                        attr.validate(value(parameter=name, **self.validation_kwargs.copy()))
                    except ValueError as e:
                        raise ValueError('Can not set attribute %s. %s'%(name, str(e)))
                    else:
                        # print "Coupling parameter: %s" % name
                        attr.couple_parameter(value)
                elif isinstance(attr, NumericParameter):
                    attr.value=value
                else:
                    raise AttributeError("Setting attribute %s is not allowed"%name)
            else:
                if self._is_protected(attr):
                    raise AttributeError("Setting attribute %s is not allowed"%name)
                else:
                    object.__setattr__(self, name, value)

    def __call__(self, **kwargs):
        """A call to the object. Will try to look up parameter as a member.

        Parameters:
            parameter(string): Name of parameter to look up

        """
        if 'parameter' not in kwargs:
            raise TypeError('Expected argument "parameter" to of a HasParameters object')

        return self.__getattribute__(kwargs['parameter'])(**kwargs)

    def _is_protected(self, attr):
        """Check if an attribute is marked as protected (implements variable protected and set to True)"""
        try:
            protected=attr.__getattribute__("protected")
        except AttributeError:
            protected=False
        return protected

    def get_parameter_list(self, numeric_types_only=False):
        """Returns a list of all the Parameters that exists

        Members that are "private", the name starting with a "_" will not be listed.

        Parameters:
            numeric_types_only (bool): if true only returns NumericParameter types.


        Returns:
            par_list (list): A list of strings with names of all the parameters.
        """
        allowed_parameter_classes=NumericParameter if numeric_types_only else Parameter
        # Loop thorough all member objects to find the allowed parameters.
        #  Note that if it is not included if it inherits HasParameters or beginning with a _
        members=self.__dict__
        par_list=[name for name in members
                  if isinstance(members[name], allowed_parameter_classes)
                  and not isinstance(members[name], HasParameters) and name[0]!='_'
                  ]
        # Loop through all the objects that inherit from HasParameter classes and add the full parameter name
        [[par_list.append(name+par_name) for par_name in
          members[name].get_parameter_list(numeric_types_only=numeric_types_only)
          ]
         for name in members if isinstance(members[name], HasParameters)
         ]
        return par_list

class Var(ArithmeticParameter):
    """ A model specific variable that will be included in the arguments to the evaluation."""

    def __init__(self, name, default_val, help=""):
        ArithmeticParameter.__init__(self, help)
        self.name=name
        self.default_val=default_val

    def _get_value(self, **kwargs):
        try:
            return kwargs[self.name]
        except KeyError:
            return self.default_val

class Calc(ArithmeticParameter):
    """A calculation done by a function of kwargs that is used as a parameter"""

    def __init__(self, function, help=""):
        ArithmeticParameter.__init__(self, help)
        self.function=function

    def _get_value(self, **kwargs):
        return self.function(**kwargs)

class Func(Calc, HasParameters):
    """ Simulates a function of parameters"""

    def __init__(self, **kwargs):
        HasParameters.__init__(self, **kwargs)

    def function(self, **kwargs):
        raise NotImplementedError("Method function in Func object not implemented.")

class Wrap(object):
    """ Defines a function where the argument can consist of parameters."""

    def __init__(self, function):
        self.function=function

    def __call__(self, *args):
        def wrap(**kwargs):
            new_args=[]
            for arg in args:
                if is_parameter(arg):
                    new_args.append(arg(**kwargs))
                else:
                    new_args.append(arg)
            return self.function(*new_args)

        wrap.__name__=self.function.__name__

        return Calc(wrap, help=wrap.__name__)

class Int(NumericParameter):
    """An integer parameter"""

    def validate(self, value):
        try:
            value=int(value)
        except Exception:
            raise ValueError('%s can not be cast to int'%value.__repr__())
        return value

class Float(NumericParameter):
    """An floating point parameter"""

    def validate(self, value):
        try:
            value=float(value)
        except Exception:
            raise ValueError('%s can not be cast to float'%value.__repr__())
        return value

class Complex(NumericParameter, HasParameters):
    """A complex number parameter"""

    def __init__(self, value, **kwargs):
        self.real=Float(0)
        self.imag=Float(0)
        NumericParameter.__init__(self, value, **kwargs)

    def validate(self, value):
        try:
            value=complex(value)
        except Exception:
            raise ValueError('%s can not be cast to complex'%value.__repr__())
        return value

    def __setattr__(self, key, value):
        if key=='value':
            value=self.validate(value)
            self.real.value=value.real
            self.imag.value=value.imag
        else:
            HasParameters.__setattr__(self, key, value)

    def _get_value(self, **kwargs):
        """Callback for getting the value of the parameter"""
        if self.has_coupled_parameter():
            return self.validate(self._coupled_parameter(**kwargs))
        else:
            return self.real(**kwargs)+1.0J*self.imag(**kwargs)

class FloatArray(NumericParameter):
    """A float array parameter"""

    # TODO: Add an array interface to support slicing as well...
    def validate(self, value):
        try:
            value=np.array(value, dtype=np.float64)
        except Exception:
            raise ValueError('%s can not be cast to float'%value.__repr__())
        return value

class ComplexArray(Complex):
    # TODO: Add an array interface to support slicing as well...
    """A complex array parameter"""

    def validate(self, value):
        try:
            value=np.array(value, dtype=np.complex128)
        except Exception:
            raise ValueError('%s can not be cast to complex array'%value.__repr__())
        return value

def get_parameters(obj, numeric_types_only=False, group_parameters=True):
    """Searches obj for members that are inherited from Parameters or HasParameters and places them in a dictonary.

    Parameters:
        obj: An object that has __dict__ (as all classes has.
        numeric_types_only (bool): returns only NumericParameters.
        group_parameters (bool): Flag to control if parameters should be grouped by type, default True.

    Returns:
        par_dict (dict): A Dictonary with class names as keys and items as lists of parameter names.
    """
    par_dict={}

    def add_parameter_to_dict(d, key, item):
        try:
            d[key].append(item)
        except KeyError:
            d[key]=[item]

    def add_object_to_dict(d, cls, name, object):
        try:
            dcls=d[cls]
        except KeyError:
            d[cls]={}
            dcls=d[cls]

        dcls[name]=object

    members=obj.__dict__
    allowed_parameter_classes=NumericParameter if numeric_types_only else Parameter
    # Loop thorough all member objects to find the allowed parameters. Note that if it is not included if it inherits
    # HasParameters
    [add_parameter_to_dict(par_dict, members[name].__class__.__name__, name)
     for name in members
     if name[0]!='_' and
     isinstance(members[name], allowed_parameter_classes) and
     not isinstance(members[name], HasParameters)
     ]
    # Loop through all the objects that inherit from HasParameter classes and do the same thing
    [add_object_to_dict(par_dict, members[name].__class__.__name__, name,
                        members[name].get_parameter_list(numeric_types_only=numeric_types_only)
                        )
     for name in members
     if isinstance(members[name], HasParameters)
     ]
    return par_dict



"""
Plotting widget for sample SLDs.
"""
import colorsys
from dataclasses import dataclass

import wx
import numpy as np
import time
import os

from scipy.interpolate import interp1d

from .reflectivity_misc import ReflectivityModule
from genx.core.custom_logging import iprint
from genx.data import DataList
from genx.gui.plotpanel import BasePlotConfig, PlotPanel
from genx.model import Model


@dataclass
class SamplePlotConfig(BasePlotConfig):
    section = 'sample plot'
    data_derived_color: bool = False
    show_single_model: bool = False
    legend_outside: bool = False

    show_imag: bool = False


class SamplePlotPanel(PlotPanel):
    '''
    Widget for plotting the scattering length density of a sample.
    '''
    opt: SamplePlotConfig

    def __init__(self, parent, plugin, id=-1, color=None, dpi=None
                 , style=wx.NO_FULL_REPAINT_ON_RESIZE, **kwargs):
        PlotPanel.__init__(self, parent, id, color, dpi, SamplePlotConfig, style, **kwargs)
        self.plugin = plugin

        self.update = self.Plot
        self.create_axes()

    def create_axes(self):
        self.ax = self.figure.add_subplot(111)
        box = self.ax.get_position()
        if self.opt.legend_outside:
            self.ax.set_position([box.x0, box.y0, box.width*0.8, box.height])
        else:
            self.ax.set_position([box.x0, box.y0, box.width*0.95, box.height])

    def Plot(self):
        '''
        Plotting the sample.
        '''
        if not self.ax:
            # self.ax = self.figure.add_subplot(111)
            self.create_axes()

        data: DataList = self.plugin.GetModel().get_data()
        if self.opt.data_derived_color:
            colors = []
            for di in data:
                h, s, v = colorsys.rgb_to_hsv(*di.sim_color)
                ci = [
                    colorsys.hsv_to_rgb(h, 1.0, 1.0),
                    colorsys.hsv_to_rgb(h, 1.0, 0.5),
                    colorsys.hsv_to_rgb(h, 0.4, 1.0),
                    colorsys.hsv_to_rgb((h+0.1)%1.0, 1.0, 1.0),
                    colorsys.hsv_to_rgb((h+0.1)%1.0, 1.0, 0.5),
                    colorsys.hsv_to_rgb((h+0.1)%1.0, 0.4, 1.0),
                    colorsys.hsv_to_rgb((h-0.1)%1.0, 1.0, 1.0),
                    colorsys.hsv_to_rgb((h-0.1)%1.0, 1.0, 0.5),
                    colorsys.hsv_to_rgb((h-0.1)%1.0, 0.4, 1.0),
                    ]
                colors.append(ci)
        else:
            colors = [['b', 'r', 'g', 'c', 'm', 'y', 'k']]*len(data)

        model: ReflectivityModule = self.plugin.GetModel().script_module
        self.plot_dicts = []
        while len(self.ax.lines)>0:
            self.ax.lines[0].remove()
        while len(self.ax.collections)>0:
            self.ax.collections[0].remove()
        i = 0
        sld_units = []

        if self.plugin.sim_returns_sld and model._sim:
            # New style sim function with one sld for each simulation
            self.plot_dicts = model.SLD
            for sim in range(len(self.plot_dicts)):
                if data[sim].show:
                    for key in self.plot_dicts[sim]:
                        try:
                            if key in ['z', 'SLD unit'] or (self.plot_dicts[0][key]==0).all():
                                # skip lines that are all zero to keep legend cleaner
                                continue
                        except KeyError:
                            pass
                        is_imag = key[:2]=='Im' or key[:4]=='imag'
                        if (is_imag and self.opt.show_imag) or not is_imag:
                            if self.opt.show_single_model:
                                label = key
                            else:
                                label = data[sim].name+'\n'+key
                            self.ax.plot(self.plot_dicts[sim]['z'], self.plot_dicts[sim][key],
                                         color=colors[sim][i%len(colors[sim])], label=label)

                            if 'SLD unit' in self.plot_dicts[sim]:
                                if not self.plot_dicts[sim]['SLD unit'] in sld_units:
                                    sld_units.append(self.plot_dicts[sim]['SLD unit'])
                            i += 1
                if self.opt.show_single_model:
                    break
        else:
            # Old style plotting just one sld
            if self.plugin.GetModel().compiled:
                try:
                    sample = model.sample
                except AttributeError:
                    iprint("Warning: Could not locate the sample in the model")
                    return
                plot_dict = sample.SimSLD(None, None, model.inst)
                self.plot_dicts = [plot_dict]
                for key in self.plot_dicts[0]:
                    if key in ['z', 'SLD unit'] or (self.plot_dicts[0][key]==0).all():
                        # skip lines that are all zero to keep legend cleaner
                        continue
                    is_imag = key[:2]=='Im' or key[:4]=='imag'
                    if (is_imag and self.opt.show_imag) or not is_imag:
                        label = key
                        self.ax.plot(self.plot_dicts[0]['z'], self.plot_dicts[0][key],
                                     colors[-1][i%len(colors[-1])], label=label)

                        if 'SLD unit' in self.plot_dicts[0]:
                            if not self.plot_dicts[0]['SLD unit'] in sld_units:
                                sld_units.append(self.plot_dicts[0]['SLD unit'])
                        i += 1

        if i>0:
            if self.opt.legend_outside:
                self.ax.legend(loc='center left', framealpha=0.5, bbox_to_anchor=(1, 0.5), fontsize="small", ncol=1)
            else:
                self.ax.legend(loc='upper right', framealpha=0.5, fontsize="small", ncol=1)
            sld_unit = ', '.join(sld_units)
            self.ax.yaxis.label.set_text('$\mathrm{\mathsf{SLD\,[%s]}}$'%sld_unit)
            self.ax.xaxis.label.set_text('$\mathrm{\mathsf{ z\,[\AA]}}$')
            wx.CallAfter(self.flush_plot)
            self.AutoScale()

    def SavePlotData(self, filename):
        ''' Save all the SLD profiles to file with filename.'''
        # Check so that there are a simulation to save
        try:
            self.plot_dicts
        except:
            self.plugin.ShowWarningDialog('No SLD data to save.'
                                          ' Simulate the model first and then save.')
            return
        base, ext = os.path.splitext(filename)
        if ext=='':
            ext = '.dat'
        data = self.plugin.GetModel().get_data()
        for sim in range(len(self.plot_dicts)):
            new_filename = (base+'%03d'%sim+ext)
            save_array = np.array([self.plot_dicts[sim]['z']])
            header = ' z [Å]'+' '*12
            for key in self.plot_dicts[sim]:
                if key!='z' and key!='SLD unit':
                    save_array = np.r_[save_array, [self.plot_dicts[sim][key]]]
                    header += f' {key:19s}'
            with open(new_filename, 'w', encoding='utf-8') as f:
                f.write("# File exported from GenX's Reflectivity plugin\n")
                f.write("# File created: %s\n"%time.ctime())
                f.write("# Simulated SLD for data set: %s\n"%data[sim].name)
                if 'SLD unit' in self.plot_dicts[sim]:
                    sld_unit = self.plot_dicts[sim]['SLD unit']
                    sld_unit = sld_unit.replace('10^{-6}\\AA^{-2}', '10⁻⁶ Å⁻²')
                    sld_unit = sld_unit.replace('r_{e}/\\AA^{3}', 'rₑ/Å⁻³')
                    f.write('# SLD unit: %s\n'%sld_unit)
                f.write("# Coumns: \n")
                f.write('#'+header+'\n')
                np.savetxt(f, save_array.T, fmt='%-19.12e')

    def generate_sld_distribution(self, reference_interface=0, number_sample=1000):
        """
        Create SLD confidence interval from a sample of random parameters chosen from
        the parameter errors. A reference surface defines the zero point for the comparison.
        """
        model_obj: Model = self.plugin.GetModel()
        parameters = model_obj.parameters
        model_module: ReflectivityModule = self.plugin.GetModel().script_module

        param_funcs, best_params, par_min, par_max = model_obj.get_fit_pars(use_bounds=False)
        param_edown, param_eup = np.array(parameters.get_error_pars()).T
        NP = len(param_funcs)

        initial_SLDs = model_module.SLD
        # setup empty lists for each element of the model SLD
        data = []
        zmin = 1e6
        zmax = -1e6
        for SLDi in initial_SLDs:
            data.append(dict([(key, []) for key in SLDi if key not in ['z', 'SLD unit']]))
            zmin = min(SLDi['z'].min(), zmin)
            zmax = max(SLDi['z'].max(), zmax)
        # create a general z-range for all simulations
        zrng = zmax-zmin
        z = np.arange(round(zmin-0.1*zrng), round(zmax+0.1*zrng), 1.0)

        if reference_interface!=0:
            d = model_module.sample.resolveLayerParameters()['d']
            h = np.cumsum(d)
            z -= h[reference_interface]
            for SLDi in initial_SLDs:
                SLDi['z'] -= h[reference_interface]

        for i in range(number_sample):
            srnd = np.random.randn(NP)
            dp = np.where(srnd>0., param_eup, param_edown)*abs(srnd)
            pi = best_params+dp
            for j, fj in enumerate(param_funcs):
                fj(pi[j])
            model_module.SLD = []
            model_module.Sim(model_obj.data)

            for i, SLDi in enumerate(model_module.SLD):
                if reference_interface!=0:
                    d = model_module.sample.resolveLayerParameters()['d']
                    h = np.cumsum(d)
                    SLDi['z'] -= h[reference_interface]

                for key, value in SLDi.items():
                    if key in ['z', 'SLD unit']:
                        continue
                    zfun = interp1d(SLDi['z'], value,
                                    fill_value=(value[0], value[-1]), bounds_error=False, kind='linear')
                    data[i][key].append(zfun(z))

        # calculate the SLD curve range that falls within the 1-sigma interval of 68.2% probability
        output = []
        for i, SLDi in enumerate(data):
            output.append({'z': z, 'SLD unit': initial_SLDs[i]['SLD unit']})
            for key, value in SLDi.items():
                zfun = interp1d(initial_SLDs[i]['z'], initial_SLDs[i][key],
                                fill_value=(initial_SLDs[i][key][0], initial_SLDs[i][key][-1]),
                                bounds_error=False, kind='linear')

                output[-1][key] = (np.percentile(np.array(value), 2.5, axis=0),  # lower bound 2-sigma
                                   np.percentile(np.array(value), 15.9, axis=0),  # lower bound 1-sigma
                                   zfun(z),  # best parameter line
                                   np.percentile(np.array(value), 84.1, axis=0),  # upper bound 1-sigma
                                   np.percentile(np.array(value), 97.5, axis=0))  # upper bound 2-sigma
        return output

    def PlotConfidence(self, reference_interface=0, number_sample=1000, plot_data=None):
        '''
        Plotting the sample.
        '''
        if not self.ax:
            # self.ax = self.figure.add_subplot(111)
            self.create_axes()

        data: DataList = self.plugin.GetModel().get_data()
        if self.opt.data_derived_color:
            colors = []
            for di in data:
                h, s, v = colorsys.rgb_to_hsv(*di.sim_color)
                ci = [
                    colorsys.hsv_to_rgb(h, 1.0, 1.0),
                    colorsys.hsv_to_rgb(h, 1.0, 0.5),
                    colorsys.hsv_to_rgb(h, 0.4, 1.0),
                    colorsys.hsv_to_rgb((h+0.1)%1.0, 1.0, 1.0),
                    colorsys.hsv_to_rgb((h+0.1)%1.0, 1.0, 0.5),
                    colorsys.hsv_to_rgb((h+0.1)%1.0, 0.4, 1.0),
                    colorsys.hsv_to_rgb((h-0.1)%1.0, 1.0, 1.0),
                    colorsys.hsv_to_rgb((h-0.1)%1.0, 1.0, 0.5),
                    colorsys.hsv_to_rgb((h-0.1)%1.0, 0.4, 1.0),
                    ]
                colors.append(ci)
        else:
            colors = [['b', 'r', 'g', 'c', 'm', 'y', 'k']]*len(data)

        while len(self.ax.lines)>0:
            self.ax.lines[0].remove()
        while len(self.ax.collections)>0:
            self.ax.collections[0].remove()

        if plot_data is None:
            plot_data = self.generate_sld_distribution(reference_interface, number_sample)
        i = 0
        sld_units = []

        # New style sim function with one sld for each simulation
        for sim in range(len(plot_data)):
            if data[sim].show:
                for key in plot_data[sim]:
                    try:
                        if key in ['z', 'SLD unit'] or (plot_data[0][key][2]==0).all():
                            # skip lines that are all zero to keep legend cleaner
                            continue
                    except KeyError:
                        pass
                    is_imag = key[:2]=='Im' or key[:4]=='imag'
                    if (is_imag and self.opt.show_imag) or not is_imag:
                        if self.opt.show_single_model:
                            label = key
                        else:
                            label = data[sim].name+'\n'+key
                        # 2-sigma range
                        self.ax.fill_between(plot_data[sim]['z'],
                                             plot_data[sim][key][0],
                                             plot_data[sim][key][4],
                                             color=colors[sim][i%len(colors[sim])], alpha=0.25)
                        # 1-sigma range
                        self.ax.fill_between(plot_data[sim]['z'],
                                             plot_data[sim][key][1],
                                             plot_data[sim][key][3],
                                             color=colors[sim][i%len(colors[sim])], alpha=0.5)
                        # fit result line
                        self.ax.plot(plot_data[sim]['z'], plot_data[sim][key][2],
                                     color=colors[sim][i%len(colors[sim])], label=label)

                        if 'SLD unit' in plot_data[sim]:
                            if not plot_data[sim]['SLD unit'] in sld_units:
                                sld_units.append(plot_data[sim]['SLD unit'])
                        i += 1
            if self.opt.show_single_model:
                break

        if i>0:
            if self.opt.legend_outside:
                self.ax.legend(loc='center left', framealpha=0.5, bbox_to_anchor=(1, 0.5), fontsize="small", ncol=1)
            else:
                self.ax.legend(loc='upper right', framealpha=0.5, fontsize="small", ncol=1)
            sld_unit = ', '.join(sld_units)
            self.ax.yaxis.label.set_text('$\mathrm{\mathsf{SLD\,[%s]}}$'%sld_unit)
            self.ax.xaxis.label.set_text('$\mathrm{\mathsf{ z\,[\AA]}}$')
            wx.CallAfter(self.flush_plot)
            self.AutoScale()

    def SaveConfidenceData(self, filename, reference_interface=0, number_sample=1000, do_plot=False):
        ''' Save all the SLD confidence intervals to file with filename.'''
        plot_data = self.generate_sld_distribution(reference_interface, number_sample)
        base, ext = os.path.splitext(filename)
        if ext=='':
            ext = '.dat'
        data = self.plugin.GetModel().get_data()
        sub_names = ['-l95', '-l63', '-bst', '-h63', '-h95']
        for sim in range(len(plot_data)):
            new_filename = (base+'%03d'%sim+ext)
            save_array = np.array([plot_data[sim]['z']])
            header = ' z [Å]'+' '*12
            for key in plot_data[sim]:
                if key!='z' and key!='SLD unit':
                    for i, sn in enumerate(sub_names):
                        save_array = np.r_[save_array, [plot_data[sim][key][i]]]
                        header += f' {key+sn:19s}'
            with open(new_filename, 'w', encoding='utf-8') as f:
                f.write("# File exported from GenX's Reflectivity plugin\n")
                f.write("# File created: %s\n"%time.ctime())
                f.write("# Simulated SLD with confidence intervals for data set: %s\n"%data[sim].name)
                f.write("# Reference interface/Number of Samples: %i/%i\n"%(reference_interface, number_sample))
                f.write("# Data describes confidence intervals given as 2-sigma (95%%), "
                        "1-sigma (68%%) and best estimate from fit.")
                if 'SLD unit' in plot_data[sim]:
                    sld_unit = plot_data[sim]['SLD unit']
                    sld_unit = sld_unit.replace('10^{-6}\\AA^{-2}', '10⁻⁶ Å⁻²')
                    sld_unit = sld_unit.replace('r_{e}/\\AA^{3}', 'rₑ/Å⁻³')
                    f.write('# SLD unit: %s\n'%sld_unit)
                f.write("# Coumns: \n")
                f.write('#'+header+'\n')
                np.savetxt(f, save_array.T, fmt='%-19.12e')
        if do_plot:
            self.PlotConfidence(plot_data=plot_data)

    def OnContextMenu(self, event):
        '''
        Callback to show the popmenu for the plot which allows various
        settings to be made.
        '''
        menu = self.generate_context_menu()
        menu.AppendSeparator()

        legendID = wx.NewId()
        menu.AppendCheckItem(legendID, "Legend Outside")
        menu.Check(legendID, self.opt.legend_outside)

        def OnLegend(event):
            self.opt.legend_outside = not self.opt.legend_outside
            self.figure.clear()
            self.create_axes()
            self.Plot()

        self.Bind(wx.EVT_MENU, OnLegend, id=legendID)

        coloringID = wx.NewId()
        menu.AppendCheckItem(coloringID, "Color from Data")
        menu.Check(coloringID, self.opt.data_derived_color)

        def OnColoring(event):
            self.opt.data_derived_color = not self.opt.data_derived_color
            self.Plot()

        self.Bind(wx.EVT_MENU, OnColoring, id=coloringID)

        singlemodelID = wx.NewId()
        menu.AppendCheckItem(singlemodelID, "Single Sample Model")
        menu.Check(singlemodelID, self.opt.show_single_model)

        def OnSingleModel(event):
            self.opt.show_single_model = not self.opt.show_single_model
            self.Plot()

        self.Bind(wx.EVT_MENU, OnSingleModel, id=singlemodelID)

        # Time to show the menu
        self.PopupMenu(menu)
        self.WriteConfig()

        self.Unbind(wx.EVT_MENU)
        menu.Destroy()
